{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvCCArray.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvCCArray;

interface
Uses SysUtils,Classes, JvCCRecrd;
{$I JCL.INC}
{$IFDEF RL_D4}
{$DEFINE DYN_ARRAY_TYPES}
{$DEFINE DYN_ARRAYS}
{$ENDIF}
const
  kMaxIntArraySize = MAXINT div sizeof(INTEGER) - 1;
  kMaxDblArraySize = MAXINT div sizeof(double) - 1;
  kIndexRangeError = 'JvCCArray: Index out of range';
{$IFDEF DYN_ARRAYS}
  kArraySize = sizeof(integer);
  kArrayRefCountSize = sizeof(integer);
  kDynArrayInfoSize =kArrayRefCountSize+kArraySize;
{$ELSE}
  kArrayHighSize = 0;
  kDynArrayInfoSize = 0;
{$ENDIF}

{$IFNDEF WIN32}
  kMaxMemBlock = 65533;
{$ENDIF}

type
  PByte = ^Byte;
  PWord = ^WORD;
  PDouble = ^Double;
  PInteger = ^Integer;
  PLongInt = ^LongInt;

  EJvCCArray = class(EJvCCRecordMgr);  {for general messages}
  EJvCCDuplicate = class(EJvCCRecordMgr);

  EJvCCArrayIndexError = class(EJvCCArray)  {for index out of range}
    constructor Create(ix:integer);
  end;

  TJvCCArray = class;
  TJvCCArrayIterator = class(TObject)
  protected
    FArray : TJvCCArray;
    FIndex : integer;
    procedure ItemsDeleted(ix,ct:integer);
    procedure ItemsInserted(ix,ct:integer);
  public
    function Init(inArray : TJvCCArray; ix : integer) : boolean;
    function NextIndex(var ix : integer) : boolean;
  end;

  TBlockHdl = ^Pointer;  {Points to the typed array}

  TJvCCSize = integer; 

  TJvCCArray = class(TJvCCBase)
  private
    FInitItemBlock : boolean; {private temporary init item block}
    FSorted : boolean; 		{false by default}
    FExactCount : boolean;	{size to exact count}
    FUniqueSort : boolean;      {throw exception on non-unique insertsort}
    FDimensions : PInteger;   {pointer to dimensions block}
    FDimPower : PInteger;   {pointer to dimension powers}
    FBlock : Pointer;       {pointer to zeroth item of array}
    FEndBlock : Pointer;        {pointer to (count - 1) item of array}
    FBlockHdl : TBlockHdl;  {pointer to another (typed?) pointer for zeroth item}
  protected
    FCapacity : TJvCCSize;   {total memory allocation of array storage}
    FDelta : integer;       {user or auto-adjustable delta}
    FIterList : TList;      {list of iterators}

    function GetItemSize : integer;
    procedure AddIterator(iter:TJvCCArrayIterator);
    procedure RemoveIterator(iter:TJvCCArrayIterator);

    procedure SetCount(inCount : TJvCCSize); override;
    procedure InitializeItems(ixfrom, ccount:integer); virtual;
    procedure FinalizeItems(ixfrom, dcount:integer); virtual;
    procedure SetCapacity(incapacity:TJvCCSize);  virtual;

    function GetItemPointer(ix : integer) : Pointer;
    function FindArrayIndex(
      var ix : integer;
      data : Pointer;
      iskey : boolean) : boolean;

    procedure SetHandle(p : TBlockHdl);
    procedure QSort(base:Pointer; nct : integer);
    procedure AltQSort(Left, Right: Integer);
    function GetAutoCapacity(inCount : integer) : integer; virtual;
    procedure FreeDimensions;

    {$IFDEF DYN_ARRAYS}
    {procedure SetLow(inLow:LongInt);}
    function GetLow : LongInt;
    function GetHigh : LongInt;
    procedure SetDynArrayCount(ct:integer);
    {$ENDIF DYN_ARRAYS}
    function GetDimCount : integer;
    property BlockHdl : TBlockHdl read FBlockHdl write SetHandle;
    property RecordMgr : TJvCCRecordMgr read FRecMgr;
  public
    constructor CreateMgr(recMgrClass : TJvCCRecordMgrClass); override;
    destructor Destroy; override;
    procedure ConfigDimensions(const dims : array of integer);
    function  Calc1DIndex(const dims : array of integer) : integer;
    procedure GetDimIX(const dimix : array of integer; var buf);
    procedure SetDimIX(const dimix : array of integer; const buf);

    procedure Config3D(x,y,z : integer);
    procedure Config2D(x,y : integer);

    procedure Get3D(x,y,z : integer; var buf);
    procedure Set3D(x,y,z : integer; const buf);
    procedure Get2D(x,y : integer; var buf);
    procedure Set2D(x,y : integer; const buf);

    function  FirstItem : TJvCCItemPtr;   override;
    function  LastItem : TJvCCItemPtr;   override;
    function  NextItem(pitem: TJvCCItemPtr):TJvCCItemPtr;   override;
    function  PriorItem(pitem: TJvCCItemPtr):TJvCCItemPtr;   override;
    function AppendItem(const buf) : TJvCCItemPtr; override;
    function DeleteItem(const buf) : boolean; override;

    procedure AppendFirst(const buf);  override;
    procedure AppendLast(const buf);  override;
    procedure Clear;  override;

    procedure BubbleSort; virtual;
    procedure QuickSort; virtual;

    procedure CopyArray(inArray:TJvCCArray);
    function  GetKeyIndex(var ix : integer; key:Pointer):boolean;
    function  GetItemIndex(var ix : integer; const buf):boolean;
    procedure SetItem(ix : integer; const buf);
    procedure GetItem(ix : integer; var buf);
    function  IndexOfItem(const buf) : integer;
    function  Compare(ix1, ix2 : integer):integer;
    procedure Swap(i1,i2:integer);
    procedure InsertInto(ix : integer; const buf);

    procedure CreateItemsAt(ix, incount : integer); { move items up}
    procedure DeleteItemsAt(ix, incount : integer); {delete space and move items down}
    procedure Delete(ix:Integer); {delete 1 item}
    {$IFDEF DYN_ARRAYS}
    property ArrayLow : LongInt read GetLow;
    property ArrayHigh : LongInt read GetHigh;
    {$ENDIF}
    property DimCount : integer read GetDimCount;
    property ItemSize : integer read GetItemSize;
    property Count : TJvCCSize read FCount write SetCount;
    property BlockPtr : Pointer read FBlock;
    property Delta : integer read FDelta write FDelta;
    property Sorted : boolean read FSorted write FSorted;
    property Unique : boolean read FUniqueSort write FUniqueSort;
    property ExactCount : boolean read FExactCount write FExactCount;
    property Capacity : TJvCCSize read FCapacity write SetCapacity;
    property Ptr[ix : integer] : Pointer read GetItemPointer;
end;

const
  kMaxObjectsCt = MAXINT div sizeof(TObject) - 1;
type
  TJvCCObjBuf = packed array[0..kMaxObjectsCt] of TObject;
  PCCObjBuf = ^TJvCCObjBuf;
{$IFDEF DYN_ARRAY_TYPES}
  TObjectArray = array of TObject;
{$ELSE}
  TObjectArray = ^TJvCCObjBuf;
{$ENDIF}


  TJvCCObjArray = class(TJvCCArray)
  private
    FPtr : TObjectArray;
    procedure SetKeyCompare(op:TJvCCKeyObjCompare);
    procedure SetObjCompare(op:TJvCCObjCompare);
    function  GetKeyCompare :TJvCCKeyObjCompare;
    function  GetObjCompare :TJvCCObjCompare;
    function  GetObject(ix : integer) : TObject;
    procedure SetObject(ix : integer; obj : TObject);
    function GetAutoDestroy : boolean;
    procedure SetAutoDestroy(dofree:boolean);
  public
    constructor CreateMgr(reclass:TJvCCRecordMgrClass); override;
    function FindKeyObject(data:Pointer):TObject;
    procedure AddObjectKey(a : TObject; data:Pointer);
    property AutoDestroy : boolean read GetAutoDestroy write SetAutoDestroy;
    property PArray : TObjectArray read FPtr;
    property Items : TObjectArray read FPtr write FPtr;
    property Objects[ix:integer] : TObject
             read GetObject write SetObject; default;
    property KeyCompare : TJvCCKeyObjCompare
             read GetKeyCompare write SetKeyCompare;
    property ObjCompare : TJvCCObjCompare
             read GetObjCompare write SetObjCompare;
  end;


  TJvCCDoubleBuf = packed array[0..kMaxDblArraySize] of double;
  PCCDoubleBuf = ^TJvCCDoubleBuf;

{$IFDEF DYN_ARRAY_TYPES}
  TArrayDouble = array of double;
{$ENDIF}
  TJvCCDoubleArray = class(TJvCCArray)
  private
    FPtr : PCCDoubleBuf;
{$IFDEF DYN_ARRAY_TYPES}
    function GetDynArray : TArrayDouble;
{$ENDIF}
  public
    constructor Create(inCount : integer);
    property PArray : PCCDoubleBuf read FPtr;
{$IFDEF DYN_ARRAY_TYPES}
    property DynArray : TArrayDouble read GetDynArray;
{$ENDIF}
  end;
  {DYN_ARRAY_TYPES usage not portable -
   Not defined - a.PArray^[0] := 3.0;
   D3+defined  a.Parray[0] := 3.0;
  }

  TJvCCSingleBuf = packed array[0..kMaxDblArraySize] of Single;
  PCCSingleBuf = ^TJvCCSingleBuf;

{$IFDEF DYN_ARRAY_TYPES}
  TArraySingle = array of Single;
{$ENDIF}
  TJvCCSingleArray = class(TJvCCArray)
  private
    FPtr : PCCSingleBuf;
{$IFDEF DYN_ARRAY_TYPES}
    function GetDynArray : TArraySingle;
{$ENDIF}
  public
    constructor Create(inCount : integer);
    property PArray : PCCSingleBuf read FPtr;
{$IFDEF DYN_ARRAY_TYPES}
    property DynArray : TArraySingle read GetDynArray;
{$ENDIF}
  end;
  {DYN_ARRAY_TYPES usage not portable -
   Not defined - a.PArray^[0] := 3.0;
   D3+defined  a.Parray[0] := 3.0;
  }

  TJvCCIntBuf = packed array[0..kMaxIntArraySize-1] of integer;
  PCCIntBuf = ^TJvCCIntBuf;

{$IFDEF DYN_ARRAY_TYPES}
  TArrayInt = array of integer;
{$ENDIF}
  TJvCCIntArray = class(TJvCCArray)
  protected
    FPtr : PCCIntBuf;
{$IFDEF DYN_ARRAY_TYPES}
    function GetDynArray : TArrayInt;
{$ENDIF}
  public
    constructor Create(inCount : integer);
    property PArray : PCCIntBuf read FPtr;
{$IFDEF DYN_ARRAY_TYPES}
    property DynArray : TArrayInt read GetDynArray;
{$ENDIF}
  end;
  {DYN_ARRAY_TYPES usage not portable -
   Not defined - a.PArray^[0] := 3;
   D3+defined  a.Parray[0] := 3;
  }
{$IFNDEF WIN32}
  {$DEFINE USE_PSTRING}
{$ENDIF}

  TJvCCSmallIntBuf = packed array[0..kMaxIntArraySize-1] of SmallInt;
  PCCSmallIntBuf = ^TJvCCSmallIntBuf;
{$IFDEF DYN_ARRAY_TYPES}
  TArraySmallInt = array of SmallInt;
{$ENDIF}
  TJvCCSmallIntArray = class(TJvCCArray)
  protected
    FPtr : PCCSmallIntBuf;
{$IFDEF DYN_ARRAY_TYPES}
    function GetDynArray : TArraySmallInt;
{$ENDIF}
  public
    constructor Create(inCount : integer);
    property PArray : PCCSmallIntBuf read FPtr;
{$IFDEF DYN_ARRAY_TYPES}
    property DynArray : TArraySmallInt read GetDynArray;
{$ENDIF}
  end;
  {DYN_ARRAY_TYPES usage not portable -
   Not defined - a.PArray^[0] := 3;
   D3+defined  a.Parray[0] := 3;
  }
{$IFNDEF WIN32}
  {$DEFINE USE_PSTRING}
{$ENDIF}


  TJvCCStrArray = class(TJvCCArray)
  protected
    FCaseSensitive : Boolean;

  public
    constructor Create(inCount : integer);
    procedure GetStr(ix:integer; var vs : string);
    procedure SetStr(ix:integer; const vs:string);
    property CaseSensitive : boolean read FCaseSensitive write FCaseSensitive;
  end;


  TIntStrRec = record
    ival : integer;
    sval : TJvCCString;
  end;
  PIntStrRec = ^TIntStrRec;

  {TJvCCIntStrArray: In this class, the ival is used as an index for a string}
  TJvCCIntStrArray = class(TJvCCArray)
  public
    constructor Create(inCount : integer);
    procedure GetIntStr(ix:integer; var vint:integer; var vs : string);
    procedure SetIntStr(ix:integer; vint:integer; const vs:string);
    procedure SetStrValue(idx:integer; const vs:string);
  end;

  TObjStrRec = record
    sval : TJvCCString;
    oval : TObject;
  end;

  PObjStrRec = ^TObjStrRec;
const
  kObjStrMax = MAXINT div sizeof(TObjStrRec) - 1;
type

  TObjStrBuf = array[0..kObjStrMax] of TObjStrRec;
  PObjStrBuf = ^TObjStrBuf;

  {TJvCCStrObjArray: In this class, the sval is used as an index for a object?}
  TJvCCStrObjArray = class(TJvCCArray)
  private
    FArrayPtr : PObjStrBuf;
    FCaseSensitive : boolean;
  public
    constructor Create(inCount : integer);
    procedure GetStr(ix:integer; var vs : string);
    procedure SetStr(ix:integer; const vs:string);
    procedure SetStrObject(const vs:string; vobj:Tobject);
    function  GetStrObject(const vs:string) : TObject;

    property ArrayPtr :  PObjStrBuf read FArrayPtr;
    property CaseSensitive : boolean read FCaseSensitive write FCaseSensitive;
  end;

    {TJvCCStrStrRec: The first string is an index for the second. No more '=' }
  TJvCCStrStrRec = record
    nval : TJvCCString;
    sval : TJvCCString;
  end;
  PCCStrStrRec = ^TJvCCStrStrRec;

  TJvCCStrStrArray = class(TJvCCArray)
  protected
    FCaseSensitive : boolean;
   public

    constructor Create(inCount : integer);
    procedure Assign(ssv :TJvCCStrStrArray);
    procedure GetStrStr(ix:integer; var pname,pvalue:string);
    procedure SetStrStr(ix:integer; const pname,pvalue:string);
    function  GetVariableIndex(var ix : integer; const pname:string) : boolean;
    Function  IndexOf(const pname : string) : integer;
    Procedure SetIndexValue(ix : integer; const s : string);
    Procedure SetIndexName(ix : integer; const s : string); {will need sort}
    Function  GetIndexValue(ix :Integer):string;
    Function  GetIndexName(ix : integer):string;
    procedure SetVariable(const pname, pvalue:string);
    procedure  GetVariable(const pname : string; var pvalue:string);
    function  GetValue(const vname : string) : string;
    procedure AssignValueStrings(slist : TStrings);
    procedure AppendValueStrings(slist : TStrings);
    procedure SetValueStrings(slist : TStrings);

    property Values[const vname:string] : string read GetValue write SetVariable;
    property CaseSensitive : boolean read FCaseSensitive write FCaseSensitive;
  end;



{$IFNDEF WIN32}
procedure ReallocMem16(var FBlock:pointer; blksize:LongInt);
{$ENDIF}


{.DEFINE ASSERT_P}

implementation
type
  TObjRef = ^TObject;
const

kMaxStepSize = 4096;

{$IFNDEF WIN32}
const
  kHdrMemBlock = 2;

{$IFDEF ASSERT_P}
  procedure ASSERT(cond:boolean; const msg:string);
  begin
    if not cond then
      raise EJvCCArray.Create(msg);
  end;
{$ENDIF}
  { the ReallocMem interface is different for delphi 1
    This routine aims to duplicate the functionality of
    the Delphi 2+ ReallocMem }
  procedure ReallocMem16(var FBlock:pointer; blksize:LongInt);
  var
    wp : PWord;
    oldsize : WORD;
  begin
    if blksize > kMaxMemBlock then
      raise EJvCCArray.Create('ReallocMem16: request exceeds maximum');
    if (FBlock=nil) then
    begin
      if (blksize > 0) then
      begin
        oldsize := blksize + kHdrMemBlock;
        GetMem(wp, oldsize);

        if wp = nil then
          raise EJvCCArray.Create('ReallocMem16: GetMem returned nil');
        wp^ := blksize;
        Inc(wp);
        FBlock := wp;
      end
    end else if (FBlock <> nil) then
    begin
      wp := FBlock;
      Dec(wp);
      oldsize := wp^;
      if (blksize <> oldsize) then
      begin
        if blksize = 0 then
        begin
          FreeMem(wp, oldsize+kHdrMemBlock);
          FBlock := nil;
        end else
        begin
          wp := ReallocMem(wp,oldsize+kHdrMemBlock,blksize+kHdrMemBlock);
          wp^ := blksize;
          Inc(wp);
          FBlock := wp;
        end;
      end;
    end;
  end;

  procedure Initialize(var s : TJvCCString);
  begin
    s := nil;
  end;

  procedure Finalize(var s : TJvCCString);
  begin
    if (s <> nil) then
      DisposeStr(s);
  end;

  function GetTStringsName(const s : string):string;
  var
    i : integer;
  begin
    for i := 1 to Length(s) do
    begin
      if (s[i] = '=') then
      begin
        Result := Copy(s,1,i-1);
        exit;
      end;
    end;
    Result := '';
  end;
{$ELSE}
{$DEFINE ASSERT_P}
{$ENDIF}


constructor EJvCCArrayIndexError.Create(ix:integer);
begin
  inherited Create('UArray: Index out of range '+IntToStr(ix));
end;


procedure TJvCCArray.FreeDimensions;
begin
  if (FDimensions <> nil) then
  begin
    Dec(FDimensions);
    FreeMem(FDimensions);
    FreeMem(FDimPower);
    FDimensions := nil;
  end;
end;

function TJvCCArray.GetDimCount : integer;
var
  pdim : PInteger;
begin
  if FDimensions = nil then
  begin
    Result := 1;
    exit;
  end;
  pdim := FDimensions;
  Dec(pdim);
  Result := pdim^;
end;

function  TJvCCArray.Calc1DIndex(const dims : array of integer) : integer;
var
  dcount : integer;
  pdim : PInteger;
  pix : PInteger;
  i : Integer;
  ix : integer;
begin
  dcount := High(dims) - Low(dims) + 1;
  ASSERT(dcount = DimCount,'Invalid dimensions');
  ix := 0;
  pix := @dims[0];
  pdim := FDimPower;
  for i := 0 to dcount-1 do
  begin
    ix := ix + pdim^ * pix^;
    Inc(pdim);
    Inc(pix);
  end;
  Result := ix;
end;

procedure TJvCCArray.ConfigDimensions(const dims : array of integer);
var
  dcount : integer;
  pdix : PInteger;
  pdim : PInteger;
  i : integer;
  dnum : integer;
begin
  dcount := High(dims) - Low(dims) + 1;
  FreeDimensions;
  FDimensions := AllocMem(sizeof(integer)*(dcount+1));
  FDimensions^ := dcount;
  FDimPower := AllocMem(sizeof(integer) * dcount);
  Inc(FDimensions);
  pdix := FDimensions;
  for i := 0 to dcount-1 do
  begin
    pdix^ := dims[i];
    Inc(pdix);
  end;
  pdim := FDimPower;
  Inc(pdim, dcount);
  dnum := 1;
  for i := dcount-1 downto 0 do
  begin
    dec(pdim);
    pdim^ := dnum;
    dnum := dnum * dims[i];
  end;
  ExactCount := TRUE;
  Count := dnum;
end;

procedure TJvCCArray.GetDimIX(const dimix : array of integer; var buf);
var
  ix : integer;
begin
  ix := Calc1DIndex(dimix);
  GetItem(ix,buf);
end;

procedure TJvCCArray.SetDimIX(const dimix : array of integer; const buf);
var
  ix : integer;
begin
  ix := Calc1DIndex(dimix);
  SetItem(ix,buf);
end;


procedure TJvCCArray.Config3D(x,y,z : integer);
begin
  ConfigDimensions([x,y,z]);
end;

procedure TJvCCArray.Config2D(x,y : integer);
begin
  ConfigDimensions([x,y]);
end;


procedure TJvCCArray.Get3D(x,y,z : integer; var buf);
var
  ix : integer;
  ppow : PInteger;
begin
  ppow := FDimPower;
  Inc(ppow);
  ix := FDimPower^ * x + ppow^ * y + z;
  GetItem(ix,buf);
end;

procedure TJvCCArray.Set3D(x,y,z : integer; const buf);
var
  ix : integer;
  ppow : PInteger;
begin
  ppow := FDimPower;
  Inc(ppow);
  ix := FDimPower^ * x + ppow^ * y + z;
  SetItem(ix,buf);
end;

procedure TJvCCArray.Get2D(x,y : integer; var buf);
var
  ix : integer;
begin
  ix := FDimPower^ * x + y;
  GetItem(ix,buf);
end;

procedure TJvCCArray.Set2D(x,y : integer; const buf);
var
  ix : integer;
begin
  ix := FDimPower^ * x + y;
  SetItem(ix,buf);
end;


procedure TJvCCArrayIterator.ItemsDeleted(ix,ct:integer);
begin
  {items starting at ix have been deleted}
  {adjust FIndex so that the item returned will be the same, or next available}
  if (FIndex >= ix) then
  begin
    if (FIndex >= ix + ct) then
      FIndex := FIndex - ct
    else {deleted what was next}
      FIndex := ix;
  end;
end;

procedure TJvCCArrayIterator.ItemsInserted(ix,ct:integer);
begin
  if (FIndex >= ix) then
  begin
    FIndex := FIndex + ct;
  end;
end;


function TJvCCArrayIterator.Init(inArray : TJvCCArray; ix : integer) : boolean;
begin       {array and index to start on, ascending}
  FArray := inArray;
  FIndex := ix;
  Result := (ix < FArray.Count);
end;

function TJvCCArrayIterator.NextIndex(var ix : integer) : boolean;
begin      {return current index, then increment}
  Result := (FIndex < FArray.Count);
  if Result then
  begin
    ix := FIndex;
    Inc(FIndex);
  end;
end;

procedure TJvCCArray.Clear;
begin
  SetCount(-1);
end;

{$IFDEF DYN_ARRAYS}
procedure TJvCCArray.SetDynArrayCount(ct:integer);
var
  pli : PLongInt;
begin
  if FBlock <> nil then
  begin
    pli := FBlock;
    Dec(pli,1);
    pli^ := Count; 
  end;
end;

function TJvCCArray.GetLow : LongInt;
begin
  Result := 0; {always 0}
end;

function TJvCCArray.GetHigh : LongInt;
var
  pi : PLongInt;
begin
  pi := FBlock;
  Dec(pi,1);
  Result := pi^ - 1;
end;
{$ENDIF}

constructor TJvCCArray.CreateMgr(recMgrClass : TJvCCRecordMgrClass);
begin
  inherited CreateMgr(recMgrClass);
  FExactCount := FALSE;
  FCapacity := -1;
  FDelta := 8;
  {$IFDEF DYN_ARRAYS}
  SetDynArrayCount(0);
  {$ENDIF}
end;

destructor TJvCCArray.Destroy;
begin
  FreeDimensions;
  inherited;
end;


procedure TJvCCArray.AddIterator(iter:TJvCCArrayIterator);
begin
  if FIterList=nil then
    FIterList := TList.Create;
  FIterList.Add(iter);
end;

procedure TJvCCArray.RemoveIterator(iter:TJvCCArrayIterator);
begin
  if FIterList = nil then
    exit;
  FIterList.Remove(iter);
end;

procedure TJvCCArray.SetHandle(p : TBlockHdl);
begin
  FBlockHdl := p;
end;

function TJvCCArray.Compare(ix1, ix2 : integer):integer;
var
  p1, p2 : PByte;
  rsize : integer;
begin
{$IFDEF ASSERT_P}
  ASSERT( ((ix1 >= 0) AND (ix1 < FCOUNT)), 'Compare range error');
  ASSERT( ((ix2 >= 0) AND (ix2 < FCOUNT)), 'Compare range error');
{$ENDIF}
  p1 := PByte(FBlock);
  p2 := p1;
  rsize := FRecMgr.ItemSize;
  Inc(p1,ix1 * rsize);
  Inc(p2,ix2 * rsize);
  Result := FRecMgr.CompareItems(p1^, p2^);
end;


procedure TJvCCArray.CopyArray(inArray:TJvCCArray);
var
  i : integer;
begin
  if (inArray.FRecMgr.ClassType <> SELF.FRecMgr.ClassType) then
  begin
    raise EJvCCArray.create('Assign of different array types');
  end;
  SELF.Count := inArray.Count;
  for i := 0 to count-1 do
  begin
    FRecMgr.AssignItem(SELF.Ptr[i]^, inArray.Ptr[i]^);
  end;
end;

procedure TJvCCArray.AltQSort(Left, Right: Integer);
var
  I: Integer;
  J: Integer;
  M : integer;
begin
  if (Right > Left) then
  repeat
    I := Left;
    J := Right;
    M := (Left + Right) shr 1;
    repeat
    {compare items in relation to record at M}
    while Compare(I,M) < 0 do
      Inc(I); {FIND FIRST record greater than or equal to M}
    while Compare(J,M) > 0 do
      Dec(J);  {FIND FIRST record less than or equal to M}
    if I <= J then
    begin
      if (I<>J) then
      begin {no point in swapping if equal index}
        Swap(I,J);
        {complication, the pivot item may have been swapped}
        if M=I then
          M := J
        else if M=J then
          M := I;
      end;
      Inc(I);
      Dec(J);
    end;
    until I > J;
    if (Left < J) then
      AltQSort(Left, J);
    Left := I;
  until I >= Right;
end;

procedure TJvCCArray.QSort(base:Pointer; nct : integer);
{Translated from C code, The standard C Library, P.J.Plauger}
var
  i : Integer;
  j : Integer;
  qi: PByte;
  qj: PByte;
  qp: PByte;
  rsize : integer;
begin
  rsize := FRecMgr.ItemSize;
  while (1 < nct) do
  begin
    i := 0;
    j := nct-1;
    qi := base;
    qj := base;

    Inc(qj,j*rsize);
    qp := qj;
    while (i < j) do
    begin
      while (i < j) and (FRecMgr.CompareItems(qi^,qp^)<=0) do
      begin
        Inc(i);
        Inc(qi,rsize);
      end;
      while (i<j) and (FRecMgr.CompareItems(qp^, qj^)<=0) do
      begin
        Dec(j);
        Dec(qj,rsize);
      end;
      if (i < j) then
      begin
        FRecMgr.SwapItems(qi^,qj^);
        Inc(i);
        Inc(qi,rsize);
      end;
    end;{while}
    if (qi <> qp) then
    begin
      FRecMgr.SwapItems(qi^,qp^);
    end;
    j := nct - i - 1;
    Inc(qi,rsize);
    if (j < i) then
    begin
      if (1 < j) then
        QSort(qi, j);
      nct := i;
    end else
    begin
      if (1 < i) then
        QSort(base, i);
      base := qi;
      nct := j;
    end;
  end;
end;

procedure TJvCCArray.QuickSort;
begin
  QSort(FBlock, Count);
  (* if (Count > 1) then {need to sort}
	      AltQSort(0, Count-1); *)
  FSorted := TRUE;
end;

function TJvCCArray.IndexOfItem(const buf) : integer;
begin
  if not GetItemIndex(Result, buf) then
    Result := -1;
end;

function  TJvCCArray.GetKeyIndex(var ix : integer; key:Pointer):boolean;
begin
  Result := FindArrayIndex(ix, key, TRUE);
end;

function TJvCCArray.FindArrayIndex(
  var ix : integer;
  data : Pointer;
  iskey : boolean) : boolean;
var
  s : integer;
  e : integer;
  mid : integer;
  match : PByte;
  base :  PByte;
  cpres : integer;
  rsize : integer;

  function SubCompare : integer;
  begin
    if iskey then
      Result := FRecMgr.CompareKey(match^,data)
    else
      Result := FRecMgr.CompareItems(match^,data^);
  end;

begin
  Result := FALSE;
  s := 0;
  e := FCount-1;
  base := FBlock;
  if (base=nil) then
  begin
    Result := FALSE;
    ix := 0;
    exit;
  end;
  rsize := FRecMgr.ItemSize;
  if not FSorted then
  begin
    match := base;
    while (s <= e) do
    begin
      if  SubCompare = 0 then
      begin
        result := TRUE;
        ix := s;
        exit;
      end;
      Inc(s);
      Inc(match,rsize);
    end;
    ix := FCount;
    exit;
  end;
  if s <= e then
  repeat
    mid := (e - s) div 2 + s;
    match := base;
    Inc(match,mid*rsize);
    cpres := SubCompare;
    if (cpres < 0) then {match < data, so go high}
      s := mid+1
    else if (cpres > 0) then {match > data, so go low}
      e := mid-1
    else begin
      Result := TRUE;
      ix := mid; {exact match}
      exit;
    end;
  until(s>e);
  ix := s;  {return insert index}
end;

function TJvCCArray.GetItemIndex(var ix : integer; const buf) : boolean;
begin
  Result := FindArrayIndex(ix, @buf, FALSE);
end;

function  TJvCCArray.GetItemPointer(ix : integer) : Pointer;
begin
  Result := FBlock;
  Inc( PByte(Result), ix * FRecMgr.ItemSize);
end;


procedure  TJvCCArray.BubbleSort;
var
  ix, ixprev : integer;
begin
  ix := 1;
  while (ix < Count)  do
  begin
    ixprev := ix-1;
    if Compare(ix,ixprev) < 0 then
    begin
      Swap(ix,ixprev);
      if (ix > 1) then
        Dec(ix);
    end else
    begin
      Inc(ix);
    end;
  end;
  FSorted := TRUE;
end;

procedure TJvCCArray.InitializeItems(ixfrom, ccount:integer);
var
  pb : PByte;
  i : integer;
  rsize : integer;
begin
  {can do override optimization using Delphi Initialize}
  pb := PByte(FBlock);
  rsize := FRecMgr.ItemSize;
  Inc(pb,ixfrom*rsize);
  for i := 0 to ccount-1 do
  begin
    FRecMgr.ConstructItem(pb^);
    Inc(pb,rsize);
  end;
end;

procedure TJvCCArray.FinalizeItems(ixfrom, dcount:integer);
var
  pb : PByte;
  i : integer;
  rsize : integer;
begin
  {can do override optimization using Delphi Finalize}
  pb := PByte(FBlock);
  rsize := FRecMgr.ItemSize;
  Inc(pb,ixfrom*rsize);
  for i := 0 to dcount-1 do
  begin
    FRecMgr.DestructItem(pb^);
    Inc(pb,rsize);
  end;
end;

procedure TJvCCArray.InsertInto(ix : integer; const buf);
begin
  CreateItemsAt(ix, 1);
  SetItem(ix, buf);
end;

function TJvCCArray.DeleteItem(const buf) : boolean;
var
  i : integer;
begin
  Result := GetItemIndex(i, buf);
  if Result then
    DeleteItemsAt(i, 1);
end;

function TJvCCArray.AppendItem(const buf) : TJvCCItemPtr;
var
  i : integer;
  pb : PByte;
begin
  if Sorted then
  begin
    if GetItemIndex(i, buf) then
    begin
      if FUniqueSort then
        raise EJvCCDuplicate.Create('Sort item duplicate');
    end;
  end else
    i := FCount;
  InsertInto(i, buf);
  pb := PByte(FBlock);
  Inc(pb,FRecMgr.ItemSize * i);
  Result := pb;
end;

procedure TJvCCArray.AppendFirst(const buf);
begin
  InsertInto(0,buf);
end;

procedure TJvCCArray.AppendLast(const buf);
begin
  InsertInto(FCount,buf);
end;

function TJvCCArray.FirstItem:TJvCCItemPtr;
begin
  if (Count > 0) then
    Result := FBlock
  else
    Result := nil;
end;

function TJvCCArray.LastItem:TJvCCItemPtr;
begin
   if (Count > 0) then
     Result := FEndBlock
   else
     Result := nil;
end;

function  TJvCCArray.NextItem(pitem: TJvCCItemPtr):TJvCCItemPtr;
begin
  if (pitem = nil) then
  begin
    Result := nil;
    exit;
  end else
  begin
    Inc(PChar(pitem), FRecMgr.ItemSize);
    if (PChar(pitem) <= PChar(FEndBlock)) then
      Result := pitem
    else
      Result := nil;
  end;
end;

function  TJvCCArray.PriorItem(pitem: TJvCCItemPtr):TJvCCItemPtr;
begin
  if (pitem = nil) then
  begin
    Result := nil;
    exit;
  end else
  begin
    Dec(PChar(pitem), FRecMgr.ItemSize);
    if (PChar(pitem) >= PChar(FBlock)) then
      Result := pitem
    else
      Result := nil;
  end;
end;



procedure TJvCCArray.CreateItemsAt(ix, incount : integer);
var
  oldcount : integer;
  src: PByte;
  dest : PByte;
  bytesmoved : integer;
  i : integer;
  rsize : integer;
begin
  if incount <= 0 then
    exit;
{$IFDEF ASSERT_P}
  ASSERT( ((ix >= 0) AND (ix <= FCOUNT)), 'Insert range error');
{$ENDIF}
  rsize := FRecMgr.ItemSize;
  oldcount := FCount;
  FInitItemBlock := TRUE;
  SetCount(incount+FCount);  {this creates uninitialised space at end}
  FInitItemBlock := FALSE;
  src := PByte(FBlock);
  Inc(src,ix*rsize);
  if(ix <> oldcount) then
  begin
    dest := src;
    bytesmoved := (oldcount-ix)*rsize;
    Inc(dest,incount*rsize);
    Move(src^,dest^,bytesmoved);
  end;
  if FRecMgr.InitItems then  {initialise inserted items}
    InitializeItems(ix,incount);
  if FIterList <> nil then
    for i := 0 to FIterList.Count-1 do
      TJvCCArrayIterator(FIterList[i]).ItemsInserted(ix,incount);
end;

function TJvCCArray.GetItemSize : integer;
begin
  Result := FRecMgr.ItemSize;
end;

procedure TJvCCArray.DeleteItemsAt(ix, incount : integer);
var
  dest : PByte;
  src : PByte;
  bytesmoved : integer;
  i : integer;
  rsize : integer;
begin
  if (incount <= 0) then exit;
{$IFDEF ASSERT_P}
    ASSERT( ((ix >= 0) AND (ix+incount <= FCOUNT)), 'Insert range error');
{$ENDIF}
  rsize := FRecMgr.ItemSize;
  dest := PByte(FBlock);
  Inc(dest, ix*rsize);
  bytesmoved := (FCount - incount - ix)*rsize;

  if FRecMgr.InitItems then
  begin
    FinalizeItems(ix,incount);
  end;
  if bytesmoved > 0 then
  begin
    src := dest;
    Inc(src, incount*rsize);
    Move(src^,dest^,bytesmoved);
  end;
  FInitItemBlock := TRUE;
  SetCount(FCount - incount);
  FInitItemBlock := FALSE;
  if FIterList <> nil then
    for i := 0 to FIterList.Count-1 do
      TJvCCArrayIterator(FIterList[i]).ItemsDeleted(ix,incount);
end;


procedure TJvCCArray.Delete(ix:Integer);
begin
  DeleteItemsAt(ix,1);
end;


procedure TJvCCArray.SetItem(ix : integer; const buf);
var
  itemPtr:PByte;
begin
  itemPtr := FBlock;
  Inc(itemPtr,ix*FRecMgr.ItemSize);
  FRecMgr.AssignItem(itemPtr^,buf);
end;

procedure TJvCCArray.GetItem(ix : integer; var buf);
var
  itemPtr:PByte;
begin
  itemPtr := FBlock;
  Inc(itemPtr,ix*FRecMgr.ItemSize);
  FRecMgr.AssignItem(buf,itemPtr^);
end;

procedure TJvCCArray.Swap(i1,i2:integer);
var
  ip1:PByte;
  ip2:PByte;
  rsize : integer;
begin
  rsize := FRecMgr.ItemSize;
  ip1 := FBlock;
  Inc(ip1,i1*rsize);
  ip2 := FBlock;
  Inc(ip2,i2*rsize);
  FRecMgr.SwapItems(ip1^,ip2^);
end;

procedure TJvCCArray.SetCapacity(incapacity:TJvCCSize);
var
  newBlockPtr : Pointer;
  newSize : LongInt;
  rsize : integer;
{$IFNDEF WIN32}
  maxCapacity : TJvCCSize;
{$ENDIF}
begin
  rsize := FRecMgr.ItemSize;
  if (incapacity >= 0) and (incapacity < FCount) then
    {setting Capacity := 0 will always shrink unused capacity}
    incapacity := FCount;
{$IFNDEF WIN32}
{ 64K block limits for this implementation }
  maxCapacity := (LongInt(kMaxMemBlock)-kDynArrayInfoSize) div FItemSize;{reuse delta as maxcapacity}
  if (incapacity >= maxCapacity) then
  begin
    if (maxCapacity <= FCount) then
      raise EJvCCArray.create('Cannot allocate more than '+IntToStr(maxCapacity-1)+' items');
    incapacity := maxCapacity;
  end;
{$ENDIF}
  newBlockPtr := FBlock;
  if (newBlockPtr <> nil) then
  begin
    Dec(PByte(newBlockPtr),kDynArrayInfoSize); {point to spare item buffer at beginning}
  end;
  if incapacity >= 0 then
    newSize := LongInt(incapacity)*rsize+kDynArrayInfoSize
  else
    newSize := 0;
{$IFNDEF WIN32}
  ReallocMem16(newBlockPtr,newSize);
{$ELSE}
  ReallocMem(newBlockPtr,newSize);{capacity plus 1 spare}
{$ENDIF}
  if newBlockPtr <> nil then
  begin
{$IFDEF DYN_ARRAYS}
    if FBlock = nil then
      PLongInt(newBlockPtr)^ := 1; {keep delphi from removing const array}
{$ENDIF}
    Inc(PByte(newBlockPtr),kDynArrayInfoSize); {point to start of array proper}
  end;
  FBlock := newBlockPtr;
  if FBlockHdl <> nil then
    FBlockHdl^ := FBlock;
  FCapacity := incapacity;
end;

function  TJvCCArray.GetAutoCapacity(inCount : integer) : integer;
  function GetDelta(orderof : integer) : TJvCCSize;
  begin
    if orderof <= 32 then
      Result := 4
    else begin
      Result :=  1;
      while (Result < orderof) do
        Result := (Result SHL  1);
      Result := Result SHR 3;
      if Result > kMaxStepSize then
        Result := kMaxStepSize;
    end;
  end;
begin
  Result := FCapacity;
  if (inCount > FCapacity) then
  begin
    FDelta := GetDelta(inCount);
    if (FDelta = 0) then
      Result := inCount
    else
      Result := (( (inCount-1) div FDelta) + 1) * FDelta;
  end else if (FDelta > 0) then
  begin
    if (FCapacity - FDelta - (FDelta shr 1) > inCount) then
    begin
      Result := (( (inCount-1) div FDelta) + 1) * FDelta;
      FDelta := GetDelta(Result);
    end;
  end;
end;


procedure TJvCCArray.SetCount(inCount : TJvCCSize);
var
  newCapacity : TJvCCSize;
  oldCount : TJvCCSize;

begin
  if (inCount = FCount) then
    exit;
  if (inCount < 0)  then
  begin
    inCount := 0;  {no swap element when count is  0}
    newCapacity := -1;  {force memory deallocation, capacity of 0 not allowed}
  end
  else begin
    if (not FExactCount) then
      newCapacity := GetAutoCapacity(inCount)
    else
      newCapacity := inCount;
  end;
  oldCount := FCount;
  if (FRecMgr.InitItems) and (not FInitItemBlock)
    and (FCount > 0) and (inCount < FCount) then
    {free items at end of array if decreasing count, prior to reallocation}
    FinalizeItems(inCount,oldCount-incount);
  FCount := inCount; {SetCapacity must know what the count is going to be}
  try
  if (newCapacity <> FCapacity) then
    SetCapacity(newCapacity);
  FEndBlock := FBlock;
  if (FEndBlock <> nil) then
    Inc(PChar(FEndBlock),(FCount-1)*FRecMgr.ItemSize);
  except
    FCount := oldCount;
    raise;
  end;
  if (FRecMgr.InitItems) and (not FInitItemBlock)
    and (inCount > 0) and (inCount > oldCount) then
    {construct items at end of array if count increased}
    InitializeItems(oldCount, inCount-oldCount);
{$IFDEF DYN_ARRAYS}
  SetDynArrayCount(FCount);
{$ENDIF}
end;


constructor TJvCCObjArray.CreateMgr(reclass:TJvCCRecordMgrClass);
begin
  inherited CreateMgr(reclass);
  BlockHdl := @FPtr;
  Count := 0;
end;

function  TJvCCObjArray.GetObject(ix : integer) : TObject;
var
  op:TObjRef;
begin
  op := TObjRef(FBlock);
  Inc(op,ix);
  Result := op^;
end;

procedure TJvCCObjArray.SetObject(ix : integer; obj : TObject);
var
  op:TObjRef;
begin
  op := TObjRef(FBlock);
  Inc(op,ix);
  op^ := obj;
end;

procedure TJvCCObjArray.SetKeyCompare(op:TJvCCKeyObjCompare);
begin
  TJvCCObjRecMgr(FRecMgr).KeyCompare := op;
end;

procedure TJvCCObjArray.SetObjCompare(op:TJvCCObjCompare);
begin
  TJvCCObjRecMgr(FRecMgr).ObjCompare := op;
end;

function  TJvCCObjArray.GetKeyCompare :TJvCCKeyObjCompare;
begin
  Result := TJvCCObjRecMgr(FRecMgr).KeyCompare;
end;

function  TJvCCObjArray.GetObjCompare :TJvCCObjCompare;
begin
  Result := TJvCCObjRecMgr(FRecMgr).ObjCompare;
end;


function TJvCCObjArray.FindKeyObject(data:Pointer):TObject;
var
  ix : integer;
begin
  if (GetKeyIndex(ix,data)) then
    Result := Objects[ix]
  else
    Result := nil;
end;

procedure TJvCCObjArray.AddObjectKey(a : TObject; data:Pointer);
var
  ix : integer;
begin
  if (GetKeyIndex(ix,data)) then
    raise EJvCCArray.Create('cannot insert duplicate object');
  InsertInto(ix,a);
end;

function TJvCCObjArray.GetAutoDestroy : boolean;
begin
  Result := TJvCCObjRecMgr(FRecMgr).AutoDestroy;
end;

procedure TJvCCObjArray.SetAutoDestroy(dofree:boolean); 
begin
  TJvCCObjRecMgr(FRecMgr).AutoDestroy := dofree;
end;



constructor TJvCCDoubleArray.Create(inCount : integer);
begin
  CreateMgr(TJvCCDoubleMgr);
  FBlockHdl := @FPtr;
  Count := inCount;
end;

{$IFDEF DYN_ARRAY_TYPES}
function TJvCCDoubleArray.GetDynArray : TArrayDouble;
begin
  Result := TArrayDouble(FPtr);
end;
{$ENDIF}



constructor TJvCCSingleArray.Create(inCount : integer);
begin
  CreateMgr(TJvCCSingleMgr);
  FBlockHdl := @FPtr;
  Count := inCount;
end;

{$IFDEF DYN_ARRAY_TYPES}
function TJvCCSingleArray.GetDynArray : TArraySingle;
begin
  Result := TArraySingle(FPtr);
end;
{$ENDIF}
constructor TJvCCIntArray.Create(inCount : integer);
begin
  CreateMgr(TJvCCIntRecMgr);
  FBlockHdl := @FPtr;
  Capacity := inCount;
  Count := inCount;
end;

{$IFDEF DYN_ARRAY_TYPES}
function TJvCCIntArray.GetDynArray : TArrayInt;
begin
  Result := TArrayInt(FPtr);
end;

{$ENDIF}

constructor TJvCCSmallIntArray.Create(inCount : integer);
begin
  CreateMgr(TJvCCSmallIntMgr);
  FBlockHdl := @FPtr;
  Capacity := inCount;
  Count := inCount;
end;

{$IFDEF DYN_ARRAY_TYPES}
function TJvCCSmallIntArray.GetDynArray : TArraySmallInt;
begin
  Result := TArraySmallInt(FPtr);
end;
{$ENDIF}


constructor TJvCCStrArray.Create(inCount : integer);
begin
  inherited CreateMgr(TJvCCStrRecMgr);
  Count := inCount;
end;



procedure TJvCCStrArray.GetStr(ix:integer; var vs : string);
var
  p:PCCStrRec;
begin
{$IFDEF ASSERT_P}
  ASSERT( ((ix>=0)AND(ix<FCount)),kIndexRangeError);
{$ENDIF}
  p := PCCStrRec(FBlock);
  Inc(p,ix);
{$IFDEF USE_PSTRING}
  vs := (p^.nval)^;
{$ELSE}
  vs := p^.nval;
{$ENDIF}
end;


procedure TJvCCStrArray.SetStr(ix:integer; const vs:string);
var
  p:PCCStrRec;
begin
{$IFDEF ASSERT_P}
  ASSERT( ((ix>=0)AND(ix<FCount)),kIndexRangeError);
{$ENDIF}
  p := PCCStrRec(FBlock);
  Inc(p,ix);
{$IFDEF USE_PSTRING}
  AssignStr(p^.nval, vs);
{$ELSE}
  p^.nval := vs;
{$ENDIF}
end;


{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
procedure TJvCCIntStrArray.SetStrValue(idx:integer; const vs:string);
var
  i : integer;
begin
  if not GetItemIndex(i, idx) then
    CreateItemsAt(i,1);
  SetIntStr(i, idx, vs);
end;


constructor TJvCCIntStrArray.Create(inCount : integer);
begin
  inherited CreateMgr(TJvCCIntStrMgr);
  Count := inCount;
  FSorted := TRUE;
end;



procedure TJvCCIntStrArray.GetIntStr(ix:integer; var vint:integer; var vs : string);
var
  p:PIntStrRec;
begin
{$IFDEF ASSERT_P}
  ASSERT( ((ix>=0)AND(ix<FCount)),kIndexRangeError);
{$ENDIF}
  p := PIntStrRec(FBlock);
  Inc(p,ix);
{$IFDEF USE_PSTRING}
  vs := (p^.sval)^;
{$ELSE}
  vs := p^.sval;
{$ENDIF}
  vint := p^.ival;
end;



procedure TJvCCIntStrArray.SetIntStr(ix:integer; vint:integer; const vs:string);
var
	p:PIntStrRec;
begin
{$IFDEF ASSERT_P}
  ASSERT( ((ix>=0)AND(ix<FCount)),kIndexRangeError);
{$ENDIF}
  p := PIntStrRec(FBlock);
  Inc(p,ix);
{$IFDEF USE_PSTRING}
  AssignStr(p^.sval, vs);
{$ELSE}
  p^.sval := vs;
{$ENDIF}
  p^.ival := vint;
end;

{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{	TJvCCStrObjArray : just to see if to works }



constructor TJvCCStrObjArray.Create(inCount : integer);
begin
  inherited CreateMgr(TJvCCStrObjMgr);
  FBlockHdl := @FArrayPtr;
  Count := inCount;
end;

procedure TJvCCStrObjArray.GetStr(ix:integer;  var vs : string);
begin
{$IFDEF USE_PSTRING}
  vs := FArrayPtr^[ix].sval^;
{$ELSE}
  vs := FArrayPtr^[ix].sval;
{$ENDIF}
end;

procedure TJvCCStrObjArray.SetStr(ix:integer;  const vs:string);
begin
{$IFDEF USE_PSTRING}
  ASSIGNSTR(FArrayPtr^[ix].sval, vs);
{$ELSE}
  FArrayPtr^[ix].sval := vs;
{$ENDIF}
end;

procedure TJvCCStrObjArray.SetStrObject(const vs:string; vobj:Tobject);
var
  ix : integer;
  rec : TObjStrRec;
  p : PObjStrRec;
begin
  FRecMgr.ConstructItem(rec);
  try
{$IFDEF USE_PSTRING}
  ASSIGNSTR(rec.sval,vs);
{$ELSE}
    rec.sval := vs;
{$ENDIF}
    if not GetItemIndex(ix,rec) then
    begin
      CreateItemsAt(ix,1);
      p := @FArrayPtr^[ix];
{$IFDEF USE_PSTRING}
      ASSIGNSTR(p^.sval,vs);
{$ELSE}
      p^.sval := vs;
{$ENDIF}
    end else
      p := @FArrayPtr^[ix];
    p^.oval := vobj;
  finally
    FRecMgr.DestructItem(rec);
  end;
end;

function TJvCCStrObjArray.GetStrObject(const vs:string):TObject;
var
  ix : integer;
  rec : TObjStrRec;
begin
  Result := nil;
  FRecMgr.ConstructItem(rec);
  try
{$IFDEF USE_PSTRING}
    ASSIGNSTR(rec.sval,vs);
{$ELSE}
    rec.sval := vs;
{$ENDIF}
    if not GetItemIndex(ix,rec) then
      raise EJvCCArray.Create('string '+vs+' not found');
    Result := FArrayPtr^[ix].oval;
  finally
    FRecMgr.DestructItem(rec);
  end;
end;


constructor TJvCCStrStrArray.Create(inCount : integer);
begin
  inherited CreateMgr(TJvCCStrStrMgr);
  Count := inCount;
  FSorted := TRUE;
end;

procedure TJvCCStrStrArray.GetStrStr(ix:integer; var pname,pvalue:string); 
var
  p : PCCStrStrRec;
begin
{$IFDEF ASSERT_P}
  ASSERT( ((ix>=0)AND(ix<FCount)),kIndexRangeError);
{$ENDIF}
  p := PCCStrStrRec(FBlock);
  Inc(p,ix);
{$IFDEF USE_PSTRING}
  pname := (p^.nval)^;
  pvalue := (p^.sval)^;
{$ELSE}
  pname := p^.nval;
  pvalue := p^.sval;
{$ENDIF}
end;

procedure TJvCCStrStrArray.SetStrStr(ix:integer; const pname,pvalue:string);
var
  p : PCCStrStrRec;
begin
{$IFDEF ASSERT_P}
  ASSERT( ((ix>=0)AND(ix<FCount)),kIndexRangeError);
{$ENDIF}
  p := PCCStrStrRec(FBlock);
  Inc(p,ix);
{$IFDEF USE_PSTRING}
  AssignStr(p^.nval,pname);
  AssignStr(p^.sval,pvalue);
{$ELSE}
  p^.nval :=  pname;
  p^.sval :=  pvalue;
{$ENDIF}
end;

function TJvCCStrStrArray.GetVariableIndex(var ix : integer; const pname:string) : boolean;
{ GetVariableIndex : return true if ix is an index for insert,
  false if pname was found, and ix is its index }
var
  match : TJvCCStrStrRec;
begin
  FRecMgr.ConstructItem(match);
{$IFDEF USE_PSTRING}
  AssignStr(match.nval,pname);
{$ELSE}
  match.nval := pname;
{$ENDIF}
  Result := GetItemIndex(ix,match);
  FRecMgr.DestructItem(match);
end;

{use StrStrArray as sorted variable list }
procedure  TJvCCStrStrArray.SetVariable(const pname, pvalue:string);
var
  ix : integer;
  prec:PCCStrStrRec;
begin
  if not GetVariableIndex(ix,pname) then
  begin
    CreateItemsAt(ix,1);
    SetStrStr(ix,pname,pvalue);
  end else
  begin
  {exists already at ix}
    prec := PCCStrStrRec(FBlock);
    Inc(prec,ix);
{$IFDEF USE_PSTRING}
    AssignStr(prec^.sval,pvalue);
{$ELSE}
    prec^.sval := pvalue;
{$ENDIF}
  end;
end;

procedure  TJvCCStrStrArray.AssignValueStrings(slist : TStrings);
begin
  SetCount(0); {clear old}
  AppendValueStrings(slist);
end;

procedure  TJvCCStrStrArray.AppendValueStrings(slist : TStrings);
var
  i : integer;
  name : string;
  value : string;
begin
  for i := 0 to slist.Count-1 do
  begin
{$IFNDEF WIN32}
    name := GetTStringsName(slist.Strings[i]);
{$ELSE}
    name := slist.Names[i];
{$ENDIF}
    value := slist.Values[name];
    SetVariable(name,value);
  end;
end;

Function TJvCCStrStrArray.GetIndexValue(ix :Integer):string;
var
  p : PCCStrStrRec;
begin
{$IFDEF ASSERT_P}
  ASSERT( ((ix>=0)AND(ix<FCount)),kIndexRangeError);
{$ENDIF}
  p := PCCStrStrRec(FBlock);
  Inc(p,ix);
{$IFDEF USE_PSTRING}
  Result := (p^.sval)^;
{$ELSE}
  Result := p^.sval;
{$ENDIF}
end;


Procedure TJvCCStrStrArray.SetIndexValue(ix : integer; const s : string);
var
  p:PCCStrStrRec;
begin
{$IFDEF ASSERT_P}
  ASSERT( ((ix>=0)AND(ix<FCount)),kIndexRangeError);
{$ENDIF}
  p := PCCStrStrRec(FBlock);
  	Inc(p,ix);
{$IFDEF USE_PSTRING}
  AssignStr(p^.sval,s);
{$ELSE}
  p^.sval := s;
{$ENDIF}
end;

Procedure TJvCCStrStrArray.SetIndexName(ix : integer; const s : string);
var
  p : PCCStrStrRec;
begin
{$IFDEF ASSERT_P}
  ASSERT( ((ix>=0)AND(ix<FCount)),kIndexRangeError);
{$ENDIF}
  p := PCCStrStrRec(FBlock);
  Inc(p,ix);
{$IFDEF USE_PSTRING}
  AssignStr(p^.nval,s);
{$ELSE}
  p^.nval := s;
{$ENDIF}
end;

Function TJvCCStrStrArray.GetIndexName(ix :Integer):string;
var
  p : PCCStrStrRec;
begin
{$IFDEF ASSERT_P}
  ASSERT( ((ix>=0)AND(ix<FCount)),kIndexRangeError);
{$ENDIF}
  p := PCCStrStrRec(FBlock);
  Inc(p,ix);
{$IFDEF USE_PSTRING}
  Result := (p^.nval)^;
{$ELSE}
  Result := p^.nval;
{$ENDIF}
end;

procedure TJvCCStrStrArray.Assign(ssv :TJvCCStrStrArray);
var
  p : PCCStrStrRec;
  q : PCCStrStrRec;
  ix : integer;
begin
  SetCount(0);
  SetCount(ssv.Count);
  p := PCCStrStrRec(FBlock);
  q := PCCStrStrRec(ssv.BlockPtr);
  for ix := 0 to Count-1 do
  begin
  {$IFDEF USE_PSTRING}
    AssignStr(p^.nval,(q^.nval)^);
    AssignStr(p^.sval,(q^.sval)^);
  {$ELSE}
    p^.nval := q^.nval;
    p^.sval := q^.sval;
  {$ENDIF}
    Inc(p);
    Inc(q);
  end;
end;


Function TJvCCStrStrArray.IndexOf(const pname : string) : integer;
begin
  if not GetVariableIndex(Result,pname) then
    Result := -1
end;

function  TJvCCStrStrArray.GetValue(const vname : string) : string;
var
  ix : integer;
  prec:PCCStrStrRec;
begin
  if not GetVariableIndex(ix,vname) then
    raise EJvCCArray.create('Variable '+vname+' not found');
  prec := PCCStrStrRec(FBlock);
  Inc(prec,ix);
{$IFDEF USE_PSTRING}
  Result := (prec^.sval)^;
{$ELSE}
  Result := prec^.sval;
{$ENDIF}
end;

procedure TJvCCStrStrArray.GetVariable(const pname : string; var pvalue:string);
var
  ix : integer;
  prec:PCCStrStrRec;
begin
  if not GetVariableIndex(ix,pname) then
    raise EJvCCArray.create('Variable '+pname+' not found');
  prec := PCCStrStrRec(FBlock);
  Inc(prec,ix);
{$IFDEF USE_PSTRING}
  pvalue := (prec^.sval)^;
{$ELSE}
  pvalue := prec^.sval;
{$ENDIF}
end;

procedure TJvCCStrStrArray.SetValueStrings(slist : TStrings);
var
  i : integer;
  s : string;
  prec:PCCStrStrRec;
begin
  slist.clear;
  prec := PCCStrStrRec(FBlock);
 	for i := 0 to FCount-1 do
  begin
{$IFDEF USE_PSTRING}
    s := (prec^.nval)^ + '=' + (prec^.sval)^;
{$ELSE}
    s := prec^.nval + '=' + prec^.sval;
{$ENDIF}
    slist.Add(s);
    Inc(prec);
  end;
end;

end.







