unit JvFileOp;

interface
Uses Sysutils,Classes;
type
  TJvRnnFileOpMode =
    (fom_none, fom_copy, fom_delete);

  TJvRnnFileOpParams = packed record
    mode : TJvRnnFileOpMode;
    prompt : boolean;
  end;


  TJvRnnFileOpCmd = record
    params : TJvRnnFileOpParams;
    opdata : string;
  end;

  TJvRnnFileOpBatch = record
    params : TJvRnnFileOpParams;
    opfiles : TStrings;
  end;

const
  kMaxFileOpBatch = MAXINT div sizeof(TJvRnnFileOpBatch)-1;
type
  TJvRnnFileOpArray = array[0..kMaxFileOpBatch] of  TJvRnnFileOpCmd;


{eg fom_delete, false, '"c:\filename"'}
{eg fom_copy, true, '"c:\source" "c:\destination"'}

procedure ExecuteRnnFileOp(oops : array of TJvRnnFileOpCmd);
procedure ExecuteRnnOpBatch( const ops : TJvRnnFileOpBatch );

implementation
Uses JvRnnPars,JVSplatFile,JvRnnStrUt,FileCtrl;

procedure sub_docmd(const op:TJvRnnFileOpCmd);
var
  s1, d1 : string;
  destpath : string;

  p : PChar;
  ptype : TJvRnnParenType;

  function GotFileName(var s : string) : boolean;
  begin
    Result := StrNextToken(p,s,ptype);
  end;

begin
  p := StrToPChar(op.opdata);
  CASE op.params.mode of
    fom_delete: begin
      if GotFileName(d1) then
        RuboutFile(d1);
      end;
    fom_copy: begin
      if GotFileName(s1) and GotFileName(d1) then
        begin
        destpath := ExtractFilePath(d1);

        if not DirectoryExists(destpath) then
          ForceDirectories(destpath);

        DoSplatFile(s1,d1,nil);
        end;
      end;
  END;
end;

procedure ExecuteRnnOpBatch( const ops : TJvRnnFileOpBatch );
var
  bop : TJvRnnFileOpCmd;
  i : integer;
begin
  bop.params := ops.params;
  for i := 0 to ops.opfiles.count-1 do begin
    bop.opdata := ops.opfiles[i];
    sub_docmd(bop);
  end;
end;

procedure ExecuteRnnFileOp(oops : array of TJvRnnFileOpCmd);
var
  i : integer;
begin
  For i := Low(oops) to High(oops) do begin
    sub_docmd(oops[i]);
  end;
end;



end.
