{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnCroTb.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnCroTb;
interface
uses DB,DBTables,Classes,SysUtils,JvCCArray,JvRnnScBDE, JvRnnScCst;

type
  TExtraRow = (prNone, prSum, prAvg);
  TExtraColumn = (ecNone, ecSum, ecAvg);

  ECrossTab = class(Exception);

  TFieldRecord = record
    xFieldName: string;
    xFieldType: TFieldType;
    xFieldSize: word;
  end;

  TJvRnnCrossTab = class(TComponent)
  private
    xExtraRow: TExtraRow;
    yField, zField : TFieldRecord;
    FTargetDB : string;
    FSourceTbl : string;
    FDestTbl : string;

    FList : TStringList;
    FTable : TTable;
    FSource : TTable;

    FGroupFields : TStrings;

    FXFieldStr : string;
    FXFieldNames : string;

    FExtraRowText : string;
    FExtraColumn : TExtraColumn;

    FOrderNumericValues : boolean;

    function copyTblData(zFld,zTyp: string): boolean;
    procedure SetTargetDB (const inDB:string);
    function GetTargetDB : string;
    procedure setMyField(fld: string; var xFld: TFieldRecord);
    function GetYValues : boolean;
    function createCrossTable: boolean;
    procedure CreateXFieldNames;
  public
    constructor Create(AOwner:TComponent);override;
    destructor Destroy; override;

    function crossTabulate(
      const pSourceTbl,destTable: string;
      crossfield,crossfunc, colfield: string;
      groupfields : TStrings;
      extraRow: TExtraRow;
      extraColumn: TExtraColumn): boolean;

    Property TargetDBName : string read GetTargetDB write SetTargetDB;
    Property ZeroPadIntegers : boolean read FOrderNumericValues write FOrderNumericValues;
    Property ExtraRowText : string read FExtraRowText write FExtraRowText;
  end;

const
  kCrossTab_F = 'CROSS_TAB';

procedure CrossTab_F(data:Pointer); far;
{@CROSS_TAB <destination table> from <source table>
<data specifier> across <column field(s)>
group by <primary key>;}

procedure InstallBDECrossTab(bdescript : TJvRnnBDEScript);

implementation
Uses JvRnnPars,JvRnnStrUt;
{$I JCL.INC}
procedure TranslateFromDef(var xFld : TFieldRecord; fldtype : TFieldType; fldsize : integer); forward;


procedure InstallBDECrossTab(bdescript : TJvRnnBDEScript);
begin
  bdescript.SetUserOp(kCrossTab_F,@CrossTab_F);
end;

constructor TJvRnnCrossTab.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);
	FList := TStringList.Create;
end;

destructor TJvRnnCrossTab.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;


procedure  TJvRnnCrossTab.SetTargetDB(const inDB:string);
begin
	FTargetDB := inDB;
end;


function  TJvRnnCrossTab.GetTargetDB : string;
begin
   Result := FTargetDB;
end;


function TJvRnnCrossTab.getYValues: boolean;
var
  qry : TQuery;
begin

  qry := TQuery.Create(nil);
  try
    qry.DatabaseName := FTargetDB;
    with qry do
    begin
      Active := false;
      SQL.Clear;
      SQL.Add('select distinct T.'''+yField.xFieldName
              +''' from '''+FSourceTbl+''' T');
      Active := true;
	    First;
      while not EOF do
      begin
        if Fields[0].AsString <> '' then FList.Add(Fields[0].AsString);
        Next;
      end;
    end;
    Result := True;
  finally
    qry.free;
  end;
end;

procedure TranslateFromDef(var xFld : TFieldRecord; fldtype : TFieldType; fldsize : integer);
begin
	xFld.xFieldType := fldtype;
	if (fldtype = ftDate) or (fldtype = ftFloat) or
		(fldtype = ftInteger) then
		xFld.xFieldSize := 0
	else
		xFld.xFieldSize := fldsize;
end;

procedure TJvRnnCrossTab.setMyField(fld: string; var xFld: TFieldRecord);
var
	def : TFieldDef;
begin
	with FSource do
	begin
		def := FieldDefs.Find(fld);
		if def <> nil then begin
			TranslateFromDef(xFld,def.DataType, def.Size);
			xFld.xFieldName := fld;
			end
		else
			raise ECrossTab.Create('['+fld+'] field does not exist in table ['+FSource.TableName+']!');
	end;
end;

procedure TJvRnnCrossTab.CreateXFieldNames;
var
	ixg : integer;
	rec : TFieldRecord;
begin
  if FGroupfields.count <= 0 then
    raise ECrossTab.create('must have fields to group on');
  {add group by key fields}
  FxFieldStr := '';
  FXFieldNames := '';

  for ixg := 0 to FGroupfields.count-1 do
  begin
    (* get a field type for each record *)
    SetMyField(FGroupfields[ixg],rec);
    if Length(FxFieldStr) > 0 then
    begin
      FxFieldStr := FxFieldStr+', ';
      FXFieldNames := FXFieldNames + ';';
    end;

    FxFieldStr := FxFieldStr+'T."'+rec.xFieldName+'"';
    FXFieldNames := FXFieldNames + rec.xFieldName;
  end;
end;

function TJvRnnCrossTab.createCrossTable: boolean;
var
	NewIndexOptions: TIndexOptions;
	extraFieldName: string[3];
	i: integer;
	vct : integer;
	qry : TQuery;
	xFld : TFieldRecord;
begin

	case FExtraColumn of
	ecSum: extraFieldName := 'SUM';
	ecAvg: extraFieldName := 'AVG';
	else   extraFieldName := '';
	end;
	NewIndexOptions := [ixPrimary, ixUnique];

	FTable.Free;
	qry := nil;

	FTable := TTable.Create(SELF);
	TRY
	  with FTable do
		begin
      TableName := FDestTbl;
      TableType := ttDefault;
      DatabaseName := FTargetDB;
      FieldDefs.Clear;
      {add the x fields (group by) }
      for i := 0 to FGroupFields.Count-1 do
      begin
        setMyField(FGroupFields[i], xFld);
        FieldDefs.Add(FGroupFields[i],xFld.xFieldType,xFld.xFieldSize, TRUE);
      end;
      {add the cross column fields, each the same type}
      for i := 0 to FList.Count-1 do
        FieldDefs.Add(FList[i], zField.xFieldType, zField.xFieldSize, False);
      {row statistics column}
      if extraFieldName <> '' then
        FieldDefs.Add(extraFieldName, zField.xFieldType, zField.xFieldSize, False);
      IndexDefs.Clear;
      IndexDefs.Add('PrimaryIndex', FXFieldNames, NewIndexOptions);
      Result := True;
		end;
    qry := TQuery.Create(nil);
    qry.DatabaseName := FTargetDB;
    with qry do
    begin
      SQL.Clear;
      SQL.Add('select distinct '+FXFieldStr +' from '''
              +FSourceTbl+''' T');
      Active := true;
	  if RecordCount = 0 then
      exit;
	  FTable.CreateTable;
	  FTable.Open;
	  vct := FGroupfields.Count-1;
    while not EOF do
	  begin
      FTable.Append;
      for i := 0 to vct do
        FTable.Fields[i].AsString := qry.Fields[i].AsString;
      FTable.Post;
      Next;
    end; { end while }
    if xExtraRow <> prNone then
    begin
      FTable.Append;
      case xExtraRow of
        prSum: FTable.Fields[0].asString := '_Sum';
        prAvg: FTable.Fields[0].asString := '_Avg';
      end; { end case }
      FTable.Fields[1].AsString := FExtraRowText;
      FTable.Post;
    end;
  end;
  EXCEPT
    FTable.Close;
    qry.Free;
    raise;
  END;

end;

function TJvRnnCrossTab.crossTabulate(const pSourceTbl,destTable: string;
						   crossfield,crossfunc,colfield: string;
						   groupfields : TStrings;
						   extraRow: TExtraRow;
						   extraColumn: TExtraColumn): boolean;






  procedure updateColumns(extraColumnType: TExtraColumn);
  var
    i, xNo, n: integer;
	xSum: double;
	vct : integer;
  begin
	if extraColumnType = ecNone then
		exit;
    try
	  with FTable do
      begin
        Exclusive := false;
        if not Active then Active := true;
        n := FieldCount - 1;
		vct := FGroupFields.Count;
		if extraColumnType <> ecNone then Dec(n); { Limit to real data fields }
		First;
		while not EOF do
        begin
          xSum := 0;
          xNo := 0;
		  Edit;
          for i := vct to n do
          begin
            if not FTable.Fields[i].IsNull then
            begin
                xSum := xSum + FTable.Fields[i].AsFloat;
                Inc(xNo);
            end;

          end; { end for }

          case extraColumnType of
            ecAvg: if xNo > 0 then
                     FieldByName('AVG').AsFloat := xSum/xNo
                   else
					 FieldByName('AVG').Clear;
            ecSum: if xNo > 0 then
                     FieldByName('SUM').AsFloat := xSum
                   else
					 FieldByName('SUM').Clear;
		  end; { case }
          Post;
          Next;
        end; { end while }
      end; { end with }
    finally
      FTable.Active := false;
    end;        
  end;

var

  sourceAlias : string;

begin
  try
	FGroupFields := groupfields;
	
	xExtraRow := extraRow;
	FExtraColumn := extraColumn;
	
	GetAliasTableName(pSourceTbl,sourceAlias,FSourceTbl);
	if Length(sourceAlias) = 0 then
		sourceAlias := FTargetDB;

  (*  if (Pos('.',FSourceTbl)=0) then
		FSourceTbl := FSourceTbl + '.DB'; *)

	FDestTbl := destTable;

	FSource := TTable.Create(SELF);
	FSource.TableName := FSourceTbl;
	FSource.DatabaseName := sourceAlias;
	FSource.FieldDefs.Update; {read in the field definitions}

	setMyField(colfield,yfield);
	setMyField(crossfield,zField);
	CreateXFieldNames; {get the key field names}

	Result := GetYValues AND
              	copyTblData(zField.xFieldName,crossfunc);


  finally
    FList.Clear;
  end;

  if Result then updateColumns(extraColumn);
end;


function ZeroPadNumeric(var s : string; ssize : integer) : boolean;
var
	i :integer;
begin
	Result := TRUE;
	{field must be purely integer digits}
	if (Length(s)=0) then
		exit;  
	for i := 1 to Length(s)  do
	begin
		if not ((s[i] >= '0') and (s[i] <= '9')) then begin
			Result := FALSE;
			exit;
		end;
	end;
	i := StrToInt(s);
	s := Format('%*.*d',[ssize,ssize,i]);

end;

function TJvRnnCrossTab.copyTblData(zFld,zTyp: string): boolean;

  function addRow(extraRowTitle: string): boolean;
  var
	i, n: integer;
	vct : integer;
	valuesList: TJvCCDoubleArray;
	countList: TJvCCIntArray;
	rowvalue : double;
	countvalue : double;
	rowfield : TFloatField;
  begin
    Result := false;
    if xExtraRow = prNone then Exit;
    n := FList.Count-1;
	vct := FGroupfields.Count;

	valuesList := TJvCCDoubleArray.Create(n+1);
	countList := TJvCCIntArray.Create(n+1);
  try
    for i := 0 to n do
    begin
      countList.PArray[i] := 0;
      valuesList.PArray[i] := 0;
    end;

    try
      with FTable do
      begin
        DatabaseName := FTargetDB;
        tableName := FDestTbl;
        Exclusive := false;
        Active := true;
        First;

        while not EOF do
        begin
          for i := 0 to n do
			if not FieldByName(FList[i]).IsNull {or
					(FieldByName(FList[i]).AsFloat < 0)} then
            begin
              valuesList.PArray^[i] := valuesList.PArray^[i]+FieldByName(FList[i]).AsFloat;
              countList.PArray^[i] := countList.PArray^[i]+1;
            end;
          Next;
        end; { end while }
		if FindKey([extraRowTitle]) then
        begin
			Edit;
			for i := 0 to n do begin
				rowfield := TFloatField(Fields[i+vct]);
				countvalue := countList.PArray^[i];
				if (countvalue > 0) then begin
					rowvalue := valuesList.PArray^[i];
					if (xExtraRow = prAvg) then
						rowvalue := rowvalue/countvalue;
					rowfield.AsFloat := rowvalue;
				end else
					rowfield.Clear;
				Result := true;
			end;{end for i}
			Post;
		end; { end if }
      end; { end with FTable }
	finally
      FTable.Active := false;
    end;
  finally
  	valuesList.free;
    countList.free;
  end;
  end;

var
  i: integer;
  s: string;
  qry : TQuery;
{$IFDEF RL_D3}
  varray : variant;
{$ENDIF}
  vct : integer;
  iv : integer;
  created : boolean;
  zdef : TField;
  recfound : boolean;
begin
  Result := false;
  qry := TQuery.Create(nil);
  qry.DatabaseName := FSource.DatabaseName;
  created := false;
  vct := FGroupFields.Count-1;
  try
{$IFDEF RL_D3}
	  if vct > 0 then
		  varray := VarArrayCreate([0,vct],varVariant);
{$ENDIF}
    for i := 0 to FList.Count - 1 do
    with qry do
	  begin
      Active := false;
      SQL.Clear;
      if Length(zTyp) > 0 then
      begin
		    s := 'select '+FXFieldStr+', '+zTyp+'(T."'+zFld+'") from '''
                + FsourceTbl+''' T where T."'+yField.xFieldName+'" = "'+FList[i]
				+ '" GROUP BY '+FXFieldStr;
      end else
      begin
			  s := 'select distinct '+FXFieldStr+', T."'+zFld+'" from '''
               + FsourceTbl+''' T where T."'+yField.xFieldName+'" = "'+FList[i]+'"';
		  end;
		  SQL.Add(s);
      Active := true;
      if RecordCount = 0 then exit;
      if not created then begin
        {ensure correct zField.xFieldType}
        zdef := Fields[vct+1];
        TranslateFromDef(zField, zdef.Datatype, zdef.DataSize);
        createCrossTable;
        created := TRUE;
      end;
      while not EOF do
      begin
        {$IFDEF RL_D3}
        if vct > 0 then
        begin
          for iv := 0 to vct do
            varray[iv] := Fields[iv].AsVariant;
        end else
          varray :=  Fields[0].AsVariant;
        recfound := FTable.Locate(FXFieldNames, varray, [loCaseInsensitive]);
        {$ELSE}
        FTable.SetKey;
        for iv := 0 to vct do
          FTable.Fields[iv].AsString := Fields[iv].AsString;
        recfound := FTable.GoToKey;
        {$ENDIF}
			  if recfound then
        begin
          FTable.Edit;
          {$IFDEF RL_D3}
          FTable.Fields[i+vct+1].AsVariant := Fields[vct+1].AsVariant;
          {$ELSE}
          FTable.Fields[i+vct+1].AsString := Fields[vct+1].AsString;
          {$ENDIF}
          FTable.Post;
				end;
			  Next;
		  end;
    end;

  finally
	  FTable.Active := false;
    qry.free;
  end;
  case xExtraRow of
    prSum: addRow('_Sum');
    prAvg: addRow('_Avg');
  end;
	if ZeroPadIntegers then
  begin
    try
      FTable.Active := TRUE;
      FTable.First;
      while not FTable.eof do
      begin
        s := FTable.Fields[0].AsString;
        if ZeroPadNumeric(s, FTable.FieldDefs[0].Size) then
        begin
          FTable.Edit;
          FTable.Fields[0].AsString := s;
          FTable.Post;
        end;
        FTable.next;
      end;{while}
	  finally
    	FTable.Active := FALSE;
	  end;
	end;
end;

procedure CrossTab_F(data:Pointer);
const
	kADD_OR_GROUP = 'ADD or GROUP by';
var
	cmdInfo : PJvScriptCmdInfo;
	p : PChar;
	scriptor : TJvRnnBDEScript;
	token : string;
	ptype : TJvRnnParenType;
	tokentype : TJvRnnTokenType;
	dest : string;
	crossfunc : string;
	source : string;
	crossfield : string;
	colfield : string;
	groupfields : TStrings;
	crosstab : TJvRnnCrossTab;
	fieldname : string;
	extraCol : TExtraColumn;

	procedure sub_error(const s : string);
	begin
		raise ERnnBDEScript.Create('CROSS_TAB: '+s);
	end;

	function sub_next : boolean;
	begin
		Result := StrNextTokenType(p,token,ptype,tokentype);
	end;

	procedure next_token;
	begin
		if not sub_next then
			sub_error('incomplete');
	end;

	procedure sub_isparen(paren : Char);
	begin
	if (not sub_next) OR (NOT ((ptype = paren_round) and (token[1] = paren))) then
		sub_error('parenthesis');
	end;

	procedure sub_fieldname(var s : string);
	begin
	if (not sub_next) OR ( not((tokentype=tok_name)OR(tokentype=tok_quotetext)) ) then
		sub_error('field name');
	s := token;
	end;

	procedure sub_tablename(var s : string);
	begin
	if (not sub_next) OR ( not((tokentype=tok_name)OR(tokentype=tok_quotetext)) ) then
		sub_error('table name');
	s := token;
	end;

	procedure sub_iskeyword(const s : string);
	begin
	if (not sub_next) OR ( COMPARETEXT(token,s) <> 0) then
		sub_error('NO '+s);
	end;


	function sub_iscomma : boolean;
	begin
		Result := sub_next;
		if Result  AND (not((tokentype=tok_bad) and (token[1] = ','))) then
			sub_error(', expected');
	end;
begin

	cmdInfo := PJvScriptCmdInfo(data);
	scriptor := cmdInfo^.scriptor as TJvRnnBDEScript;

	p := cmdInfo^.cmdcstr;

	{get destination name}
	sub_tablename(dest);
	sub_iskeyword('FROM');
	sub_tablename(source);

	if (not sub_next) then
		sub_error('cross function');

	if COMPARETEXT(token,'VALUES')=0 then
		crossfunc := ''
	else
		crossfunc := token;

	sub_isparen('(');
	sub_fieldname(crossfield);
	sub_isparen(')');
	sub_iskeyword('ACROSS');
	sub_fieldname(colfield);

	extraCol := ecNone;

	if not sub_next then
		  sub_error(kADD_OR_GROUP);

	if COMPARETEXT('ADD',token) = 0 then begin
		if sub_next then begin
			if COMPARETEXT('AVG',token)=0 then
				extraCol := ecAvg
			else
			if COMPARETEXT('SUM',token)=0 then
				extraCol := ecSum
		end;
		if extraCol = ecNone then
			sub_error('SUM or AVG expected');

		sub_iskeyword('GROUP');
		sub_iskeyword('BY');

	end
	else
	if (COMPARETEXT('GROUP',token)=0) then
		sub_iskeyword('BY')
	else
		sub_error(kADD_OR_GROUP);


	groupfields := TStringList.Create;
try
	repeat
	sub_fieldname(fieldname);
	groupfields.Add(fieldname);
	until (not sub_iscomma);

	crosstab := TJvRnnCrossTab.Create(nil);
try
	crosstab.TargetDBName := scriptor.Alias;
	crosstab.crossTabulate(source,dest,crossfield,crossfunc,colfield,
							groupfields,prNone,extraCol);
finally
	crosstab.free;
end;
finally
	groupfields.free;
end;

end;

end.
 
