{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnScrSt.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}

unit JvRnnScrSt;  { TRyScriptStore component. }

interface


uses
  WinProcs, 
  SysUtils, 
  Messages, 
  Classes, 
  Graphics, 
  Controls, 
  Forms, 
  Dialogs, 
  Menus, 
  StdCtrls, 
  ExtCtrls,
  JvRnnStrLs,
  JvCCArray,JvCCRecrd;
  
type
  { a script list holds
  a list of labelled scripts
  }
  ERnnScript=class(Exception);
  
  TJvRnnScriptList = class;
  
  TJvRnnScript = class(TObject)
  private
    FScriptName : PString;
    FScriptType : PString;
    FStrx :  TStrings;
    FScriptList : TJvRnnScriptList;
  
    procedure SetScript(slist : TStrings);
  
  public
    constructor Create(const scrname,scrtype:string; slist:TJvRnnScriptList);
    destructor  Destroy; override;

    procedure DetachList;
    procedure SetList(slist:TJvRnnScriptList);

    procedure ReadScript(Reader:TReader);
    procedure WriteScript(Writer:TWriter);

    procedure SaveToStream(stream:TStream);

    procedure SetName(const inName:string);
    function GetName : string;

    procedure SetType(const inType:string);
    function GetType : string;

    {properties}
    property Strx : TStrings read FStrx write SetScript;
    property Name : string read GetName write SetName;
    property ScriptType : string read GetType write SetType;
  end;

  TJvRnnScriptHolder = class(TComponent)
  private
    FKey : string;
    FScript : TStrings;
    FEncryptHex : Boolean;
    procedure SetScript(slist:TStrings);
    function GetEncrypt : boolean;
    public
    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;
    procedure LoadFromFile(const fname:string);
    procedure SaveToFile(const fname:string);
    published
    property Name;
    property Tag;
    property EncryptKey : string read FKey write FKey;
    property Encrypt : boolean read GetEncrypt;
    property EncryptHex : boolean read FEncryptHex write FEncryptHex default FALSE;
    property Script : TStrings read FScript write SetScript;
  end;

  TJvRnnScriptList = class(TPersistent)
  private
    FList : TJvCCObjArray;
    procedure ReadScriptList(Reader:TReader);
    procedure WriteScriptList(Writer:TWriter);

    protected
    procedure DefineProperties(Filer: TFiler); Override;

  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure Assign(obj : TPersistent);override;
    procedure TakeOverScripts(scriptList : TJvRnnScriptList);

    procedure AddScript(const scriptName,scrtype : string; slist : TStrings);
    procedure DeleteScript(const scriptName : string);
    function  GetScript(const scriptName : string) : TJvRnnScript;
    function  GetCount : integer;
    function  GetScriptStrings(const scriptName : string) : TStrings;
  
    procedure LoadFromFile(const inName : string);
    procedure SaveToFile(const inName : string);
  
    procedure LoadFromStream(instream : TStream);
    procedure SaveToStream(outstream : TStream);

    function GetScriptNum(ix:integer) : TJvRnnScript;

    property Scripts[ix:integer] : TJvRnnScript read GetScriptNum;
    property Count : integer read GetCount;
  end;

  TJvRnnScriptListClass = class of TJvRnnScriptList;
  
  TJvRnnScriptStore = class(TComponent)
  private
    FScripts : TJvRnnScriptList;
  protected
    FShouldSaveScripts: Boolean;
  function  FetchScript(const scrlabel:string) : TStrings;
  
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure ClearScripts;
    procedure GetScriptAsList(const scrlabel:string; slist : TStrings);
    function  GetScriptAsCStr(const scrlabel:string):PChar;
    {warning!  GetScriptText returns ShortStrings in delphi 1}
    function  GetScriptText(const scrlabel:string):string;
    procedure SetScripts(inscr : TJvRnnScriptList);

  published
    property ScriptList : TJvRnnScriptList read FScripts write SetScripts;
  
  end;  { TJvRnnScriptStore }

TJvRnnScriptStoreClass = class of TJvRnnScriptStore;



implementation
Uses JvRnnStrUt,JvRnnPars,JvRnnStrms,JvRnnDupDg;
{---------------------}

function CompareScriptName(obj1,obj2:TObject):integer; forward;
function CompareKey(obj:TObject;key:Pointer):integer; forward;

{$IFNDEF WIN32}
  {$R *.R16}
{$ENDIF}
  
procedure TJvRnnScript.SetScript(slist : TStrings);
begin
  FStrx.Assign(slist);
end;
  
  
constructor TJvRnnScript.Create(const scrname,scrtype:string; slist:TJvRnnScriptList);
begin
  inherited Create;
  Name := scrname;
  ScriptType := scrtype;
  FStrx := TJvRnnStrList.Create;
  SetList(slist);
end;
  
destructor  TJvRnnScript.Destroy;
begin
  DetachList;
  FStrx.Free;
  if (FScriptName <> nil) then
    DisposeStr(FScriptName);
  if (FScriptType <> nil) then
    DisposeStr(FScriptType);
  inherited Destroy;
end;
  
  
procedure TJvRnnScript.SaveToStream(stream:TStream);
const
  kEOL : array[0..1] of Char = #13#10;
var
  outhdr : string;
begin
  {outhdr := Format('/*$ %s <%s> */',[ScriptType,Name]);}
  outhdr := Format('<script language = "%s" ID =  "%s">',[ScriptType,Name]);
  stream.Write(outhdr[1],Length(outhdr));
  stream.Write(kEOL,2);
  Strx.SaveToStream(stream);
  outhdr := '<\script>';
  stream.Write(outhdr[1],Length(outhdr));
  stream.Write(kEOL,2);
end;
  
  
procedure TJvRnnScript.SetName(const inName:string);
var
  slist : TJvRnnScriptList;
begin
  slist := FScriptList;
  DetachList;
  AssignStr(FScriptName, inName);
  if slist <> nil then
    SetList(slist);
end;
  
function TJvRnnScript.GetType : string;
begin
  if FScriptType=nil then
    Result := ''
  else
    Result := FScriptType^;
end;
  
  
procedure TJvRnnScript.SetType(const inType:string);
begin
  AssignStr(FScriptType, inType);
end;
  
function TJvRnnScript.GetName : string;
begin
  if FScriptName=nil then
    Result := ''
  else
    Result := FScriptName^;
end;
  
procedure TJvRnnScript.SetList(slist : TJvRnnScriptList);
begin
  FScriptList := slist;
  try
    FScriptList.FList.AddObjectKey(SELF,FScriptName);
  except
    on x1:EJvCCArray do
    begin
      MessageDlg(x1.message, mtError, [mbOK], 0);
    end;
  end;
end;
  
procedure TJvRnnScript.DetachList;
begin
  if FScriptList <> nil then
    FScriptList.FList.DeleteItem(SELF);
  FScriptList := nil;
end;
  
procedure TJvRnnScript.ReadScript(Reader:TReader);
begin
  FStrx.clear;
  reader.ReadListBegin;
  while not reader.EndOfList do
    FStrx.Add(reader.ReadString);
  reader.ReadListEnd;
end;
  
  
procedure TJvRnnScript.WriteScript(Writer:TWriter);
var
  i : integer;
begin
  writer.WriteString(ScriptType);
  writer.WriteListBegin;
  for i := 0 to FStrx.Count-1 do
    writer.WriteString(FStrx[i]);
  writer.WriteListEnd;
end;
  
  
constructor TJvRnnScriptHolder.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FScript := TStringList.Create;
end;
  
destructor TJvRnnScriptHolder.Destroy;
begin
  FScript.Free;
  inherited Destroy;
end;
 
procedure TJvRnnScriptHolder.SetScript(slist:TStrings);
begin
  FScript.Assign(slist);
end;
  
function TJvRnnScriptHolder.GetEncrypt : boolean;
begin
  Result := Length(FKey) > 0;
end;
  
procedure TJvRnnScriptHolder.LoadFromFile(const fname:string);
begin
  FScript.LoadFromFile(fname);
  if Encrypt then
    DecryptStrings(FKey,FScript,EncryptHex);
end;
  
procedure TJvRnnScriptHolder.SaveToFile(const fname:string);
begin
  if Encrypt then
    EncryptStrings(FKey,FScript,EncryptHex);
end;
  
constructor TJvRnnScriptList.Create;
begin
  inherited Create;
  FList := TJvCCObjArray.CreateMgr(TJvCCObjRecMgr);
  FList.ObjCompare := CompareScriptName;
  FList.KeyCompare := CompareKey;
end;
  
function  TJvRnnScriptList.GetCount : integer;
begin
  Result := FList.Count;
end;
  
function TJvRnnScriptList.GetScriptNum(ix:integer) : TJvRnnScript;
begin
  Result := TJvRnnScript(FList[ix]);
end;
  
  
procedure TJvRnnScriptList.Clear;
var
  script : TJvRnnScript;
  i : integer;
begin
  for i := 0 to FList.Count-1 do begin
  script := TJvRnnScript(FList[i]);
  script.FScriptList := nil;
  script.free;
  end;
  FList.Clear;
end;
  
function  TJvRnnScriptList.GetScript(const scriptName : string) : TJvRnnScript;
begin
  Result := TJvRnnScript(FList.FindKeyObject(@scriptName));
end;
  
function CompareScriptName(obj1, obj2 : TObject):integer;
begin
  Result := CompareText(TJvRnnScript(obj1).Name, TJvRnnScript(obj2).Name);
end;

function CompareKey(obj: TObject; key:Pointer):integer;
begin
  Result := CompareText(TJvRnnScript(obj).Name, PString(key)^);
end;

procedure TJvRnnScriptList.TakeOverScripts(scriptList : TJvRnnScriptList);
var
  temp:TJvCCObjArray;
begin
  {the scripts in FList of scriptList are assigned to SELF,
  by clearing own FList, and swapping FLists of scriptList.
  The result is that scriptlist is cleared, and SELF takes over the scripts}
  Clear;  {clear own}
  temp := FList;
  FList := scriptList.FList;  {get scripts}
  scriptList.FList := temp;   {empty list to scriptList}
end;
  
procedure TJvRnnScriptList.ReadScriptList(Reader:TReader);
var
  script : TJvRnnScript;
  sname : string;
  stype : string;
begin
   Clear;
  
   reader.ReadListBegin;
   while not reader.EndOfList do begin
    sname := reader.ReadString;
    stype := reader.ReadString;
    script := TJvRnnScript.Create(sname,stype,SELF);
    script.ReadScript(reader);
   end;
   reader.ReadListEnd;
end;
  
  
procedure TJvRnnScriptList.DefineProperties(Filer: TFiler);
begin
  Inherited DefineProperties(Filer);
  with Filer do
  begin
    DefineProperty('ScriptList', ReadScriptList, WriteScriptList, (Count > 0) );
  end;
end;
  
  
procedure TJvRnnScriptList.WriteScriptList(Writer:TWriter);
var
  script : TJvRnnScript;
  i : integer;
begin
   writer.WriteListBegin;
   for i := 0 to FList.Count-1 do begin
    script := TJvRnnScript(FList.Objects[i]);
    writer.WriteString(script.Name);
    script.WriteScript(writer);
   end;
   writer.WriteListEnd;
end;
  
  
function  TJvRnnScriptList.GetScriptStrings(const scriptName : string) : TStrings;
var
  script : TJvRnnScript;
begin
  script := GetScript(scriptName);
  if script <> nil then
    Result := script.Strx
  else
    Result := nil;
end;
  
procedure TJvRnnScriptList.SaveToStream(outstream : TStream);
var
  script : TJvRnnScript;
  ix : integer;
begin
  {will always output scripts by name}
  for ix := 0 to FList.Count-1 do begin
    script := TJvRnnScript(FList.Objects[ix]);
    script.SaveToStream(outstream);
  end;
       
end;
  
procedure TJvRnnScriptList.LoadFromFile(const inName : string);
var
  stream : TFileStream;
begin
  stream := TFileStream.Create(inName, fmOpenRead	or fmShareDenyWrite);
try
  LoadFromStream(stream);
finally
  stream.free;
end;
  
end;
  
procedure TJvRnnScriptList.SaveToFile(const inName : string);
var
  stream : TFileStream;
begin
  stream := TFileStream.Create(inName, fmOpenWrite	or fmShareExclusive);
try
  SaveToStream(stream);
finally
  stream.free;
end;
  
end;
  
  
procedure TJvRnnScriptList.LoadFromStream(instream : TStream);
const
{$IFDEF WIN32}
  kMaxLineBuf = 0;
{$ELSE}
  kMaxLineBuf = 256;
{$ENDIF}
  kIgnoreStuff = 0;
  kScriptSpec = 1;
  kEndScript = 2;
  kComScriptSpec = 3;

var
  s: string;
  linePtr : PChar;
  p : PChar;
  token : string;
  scrname, scrtype : string;
  attributes : TJvRnnStrList;

  doSkip : boolean;
  
  slist: TStrings;
  script : TJvRnnScript;
  parentype : TJvRnnParenType;
  toktype : TJvRnnTokenType;

  bscript : TJvRnnScript;
  lineread : TJvRnnStreamRead;
  specno : integer;
  exitAll : boolean;
  inxscript : boolean;
  eoline : boolean;
  eofile : boolean;

  procedure ThrowError(const s  : string);
  begin
  raise ERnnScript.create(Format('Script Extraction Failed : %s',[s]));
  end;

  procedure ThrowNameError;
  begin
  ThrowError('<script  language = "Type" ID = "Name"> expected')
  end;

  procedure AddToScript(const line : string);
  begin
  if (slist <> nil) then
  begin
    slist.Add(line);
  end;
  end;

  function GetNextLine : boolean;
  begin
  linePtr := lineread.NextLine;
  eofile := (linePtr=nil);
  Result := (not eofile);
  p := linePtr;
  eoline := (eofile) OR (p^ = #0);
  end;

  function GetNextToken : boolean;
  begin
  Result := FALSE;
  while (not Result) do
  begin
    if (eoline) then
    if not GetNextLine then
      exit;
    Result := StrNextToken(p, token, parentype);
    eoline := (p^ = #0);
  end;
  end;

  function GetNextTokenType : boolean;
  begin
  Result := FALSE;
  while (not Result) do
  begin
    if (eoline) then
    if not GetNextLine then
      exit;
    Result := StrNextTokenType(p, token, parentype,toktype);
    eoline := (p^ = #0);
  end;
  end;

  function GetSpecStuff : integer;
  begin
  Result := kIgnoreStuff;
  p := linePtr;
  if p = nil then
    exit;
  while (p^ <> #0) do
  begin
    if (p^ = '/') then
    begin
    Inc(p);
    if (p^ = '*') and (p[1] = '$') then
    begin
      Inc(p,2);
      Result := kComScriptSpec;
      break;
    end
    end else
    if (p^ = '<') then
    begin
    Inc(p);
    if StrLComp('script',p,6)=0 then
    begin
      Inc(p,6);
      Result := kScriptSpec;
      break;
    end else
    if StrLComp('\script>',p,8)=0 then
    begin
      if not inxscript then
        ThrowError('unmatched <\script>');
      SetLength(s,0);
      AddCStrSeg(s,linePtr,p-1);   {get unsaved in s}
      if (Length(s) > 0) then
        AddToScript(s);
      Inc(p,8);
      Result := kEndScript;
      break;
    end;
    end else
    Inc(p);
  end;
  if (Result <> kIgnoreStuff) then
    linePtr := p;
  end;

  procedure Sub_NoScript;
  begin
  slist := nil;
  script := nil;
  scrname := '';
  scrtype := '';
  inxscript := FALSE;
  end;
  
  procedure Sub_StoreScript;
  begin
  Sub_NoScript;
  end;

  procedure StartScript;
  begin
  { rest of current line belongs to script }

  if scrname='' then
    ThrowError('script ID expected');
  bscript := GetScript(scrname);
  doSkip := FALSE;
  while bscript <> nil do
  begin
    CASE GetDupAction(scrname) of
    dup_replace:
      begin
      bscript.free;
      bscript := nil;
      end;
    dup_rename:
      begin
      bscript := GetScript(scrname);
      end;
    dup_skip:
      begin
      bscript := nil;
      doSkip := TRUE;
      end;
    dup_abort:
      begin
      ThrowError('load from stream aborted');
      end;
    END;
  end; {while}
  if (not doSkip) then
  begin
    script := TJvRnnScript.Create(scrname, scrtype, SELF);
    slist := script.Strx;
    lineptr := p; {shorten the line to here}
  end;
  end;


  procedure DoEndScript;
  begin
  if (not inxscript) then
    ThrowError('No <script..> for <\script>');
  Sub_StoreScript;
  inxscript := FALSE;
  end;

  procedure DoScriptSpec; {get script attributes}
  begin
  if inxscript then
    ThrowError('Recursive script inclusion not allowed');
  Sub_StoreScript; {ensure new script}
  attributes.Clear;
  while GetNextTokenType do
  begin
    CASE toktype of
    tok_name :
      begin
      scrname := token;
      if (not GetNextTokenType) AND (toktype<>tok_equal) then
        ThrowError('attribute "=" expected');
      if (not GetNextTokenType) AND
        (not((toktype=tok_name)OR(toktype=tok_quotetext))) then
        ThrowError('attribute "value" expected');
      attributes.Add(Format('%s=%s',[scrname,token]));
      end;
    tok_greaterthan :
      begin
      break; {ready to start?}
      end;
    ELSE
     ThrowError(Format('Unexpected token %s',[token]));
    END;
  end;
  scrtype := attributes.Values['language'];
  scrname := attributes.Values['ID'];
  if (Length(scrname) = 0) then
    ThrowError('Missing script ID');
  if (Length(scrtype) = 0) then
    ThrowError('Missing script language');
  StartScript;
  inxscript := TRUE;
  end;

  procedure DoScriptCom;
  begin
  if inxscript then
    exit;
  Sub_StoreScript;
  if (not GetNextToken) then
    ThrowError('script type label expected');
  scrname := '';
  scrtype := token;
  if (parentype=paren_angle) then {no type given?}
  begin
    scrname := scrtype;
    scrtype := ''; {untyped}
  end else
  begin {check for special types?}
    scrname := '';
    if CompareText(scrtype,'END')=0 then
    begin
    Sub_StoreScript;
    exit;
    end;
    if CompareText(scrtype,'STOP')=0 then
    begin
    Sub_StoreScript;
    exitAll := TRUE;
    exit;
    end;
  end;
  while GetNextToken do
  begin
    if (token = '*/') then
    break;
    if (Length(scrname) = 0) and (parentype=paren_angle) then
    scrname := token;
  end;
  StartScript;
  end;

begin
  inxscript := FALSE;
  exitAll := FALSE;
  attributes := TJvRnnStrList.Create;
  TRY
  Sub_NoScript;
  lineread := TJvRnnStreamRead.Create(instream, kMaxLineBuf);
  lineread.FixedSize := TRUE; {no change in stream size during read}
  TRY
    eofile := FALSE;
    eoline := TRUE;
    while (not eofile) and (not exitAll) do
    begin
    if (eoline) then
      if not GetNextLine then
      break;
    specno := GetSpecStuff;
    CASE specno of
      kIgnoreStuff :
      begin
        AddToScript(StrPas(linePtr));
        eoline := TRUE;
      end;
      kComScriptSpec :
      DoScriptCom;
      kScriptSpec :
      DoScriptSpec;
      kEndScript :
      DoEndScript;
    END;
    lineptr := p;
    eoline := (p=nil) OR (p^ = #0);
    end; {while}
    if inxscript then
    ThrowError('Missing <\script> tag');
    Sub_StoreScript; {end last}
  FINALLY
    lineread.free;
  END;
  FINALLY
  attributes.free;
  END;
end;

destructor TJvRnnScriptList.Destroy;
var
  i : integer;
  script : TJvRnnScript;
begin
  for i := Count-1 downto 0 do
  begin
    script := TJvRnnScript(FList[i]);
    script.FScriptList := nil;
    FList.Delete(i);
    script.free;
  end;
  FList.Free;
  inherited Destroy;
end;

procedure TJvRnnScriptList.Assign(obj : TPersistent);
var
  i : integer;
  script : TJvRnnScript;
  scriptList : TJvRnnScriptList;
begin
  if obj is TJvRnnScriptList then begin
    scriptList := TJvRnnScriptList(obj);
    Clear;
    for i := 0 to scriptList.Count-1 do  begin
      script := scriptList.Scripts[i];
      AddScript(script.Name, script.ScriptType, script.Strx);
    end;
  end
  else
    inherited Assign(obj);
end;
  
procedure TJvRnnScriptList.AddScript(const scriptName,scrtype : string; slist : TStrings);
var
  script : TJvRnnScript;
begin
  script := TJvRnnScript.Create(scriptName,scrtype,self);
  script.Strx := slist;
end;

procedure TJvRnnScriptList.DeleteScript(const scriptName : string);
var
  script : TJvRnnScript;
begin
  script := GetScript(scriptName);
  script.Free;
end;

{---------------------}




{
TJvRnnScriptStore
}
function  TJvRnnScriptStore.FetchScript(const scrlabel:string) : TStrings;
begin
  Result := FScripts.GetScriptStrings(scrlabel);
end;

function  TJvRnnScriptStore.GetScriptText(const scrlabel:string):string;
var
  script : TStrings;
begin
  script := FetchScript(scrlabel);
  Result := StrListToString(script,' ');
end;

function   TJvRnnScriptStore.GetScriptAsCStr(const scrlabel:string):PChar;
var
  script : TStrings;
begin
  script := FetchScript(scrlabel);
  Result := StrListToCStr(script,' ');
end;


procedure TJvRnnScriptStore.GetScriptAsList(const scrlabel:string; slist : TStrings);  { public }
var
  script : TStrings;
begin
  script := FetchScript(scrlabel);
  slist.Assign(script);
end;  { GeTJvRnnScriptAsList }
  
procedure TJvRnnScriptStore.ClearScripts;
begin
  FScripts.Clear;
end;
  
  
procedure TJvRnnScriptStore.SetScripts(inscr : TJvRnnScriptList);
begin
  FScripts.Assign(inscr);
end;
  
(*
procedure TJvRnnScriptStore.ReadScripts(Reader: TReader);
begin
  FScripts.ReadScriptList(Reader);
end;
  
procedure TJvRnnScriptStore.WriteScripts(Writer: TWriter);
begin
  FScripts.WriteScriptList(Writer);
end;
  
procedure TJvRnnScriptStore.DefineProperties(Filer: TFiler);
begin
  Inherited DefineProperties(Filer);
  if FShouldSaveScripts then
  with Filer do
  begin
    DefineProperty('Scripts', ReadScripts, WriteScripts, True {Has Data} );
  end;
end;*)
  
destructor TJvRnnScriptStore.Destroy;
begin
  { CDK: Free allocated memory and created objects here. }
  { Free member variables: }
  FScripts.Free;
  
  inherited Destroy;
end;  { Destroy }
  
constructor TJvRnnScriptStore.Create(AOwner: TComponent);
{ Creates an object of type TJvRnnScriptStore, and initializes properties. }
begin
  inherited Create(AOwner);
  { Create member variables (that are objects): }
  FScripts := TJvRnnScriptList.Create;
  FShouldSaveScripts := True;
  
end;  { Create }
  
  
initialization
  RegisterClasses([TJvRnnScriptList,TJvRnnScriptStore]);
end.
