{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnStrLs.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnStrLs;

interface
Uses classes, Sysutils, JvCCArray;
{$I JCL.INC}

type
  ERnnStrList = class(Exception);

  TJvRnnStrList = class(TStrings)
  private
    FStrArray: TJvCCStrObjArray;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    FSorted : Boolean;
    FDuplicates : TDuplicates;
    procedure SetSorted(Value: Boolean);
    procedure InsertItem(Index: Integer; const S: string);
  protected
    procedure Changed; virtual;
    procedure Changing; virtual;
    function Get(Index: Integer): string; override;
  {$IFNDEF VER80}
    function GetCapacity: Integer; override;
    procedure SetCapacity(NewCapacity: Integer); override;
  {$ENDIF}
    function GetCount: Integer; override;
    procedure Put(Index: Integer; const S: string); override;
    procedure SetUpdateState(Updating: Boolean); override;
  {$IFDEF VER80}
    function GetStrName(ix:integer):string;
  {$ENDIF}
  public
    constructor Create;
    destructor Destroy; override;
    function Add(const S: string): Integer; override;
    procedure Clear; override;
    function IndexOf(const S: string): Integer; override;
    function Find(const S: string; var Index: Integer): Boolean;
    procedure PutObject(ix : integer; vobj : Tobject); override;
    function  GetObject(ix : integer) : TObject; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure Sort; virtual;

    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
    property StrArray : TJvCCStrObjArray read FStrArray; {got to have this}
    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    property Sorted: Boolean read FSorted write SetSorted;
  {$IFDEF VER80}
    property Names[ix:integer] : string read GetStrName;
  {$ENDIF}
  end;
implementation
Uses Consts,JvRnnStrUt;
{$IFDEF VER80}
const
    SListIndexError =  'Index %d out of range';
    SSortedListError = 'Insert at %d when sorted';
    SDuplicateString = 'Duplicate at %d';
    
procedure Error(const errmsg : string; atindex : integer);
var
  s : string;
begin
  s := Format(errmsg, [atindex]);
  raise ERnnStrList.Create(s);
end;
{$ENDIF}

destructor TJvRnnStrList.Destroy;
begin
  FOnChange := nil;
  FOnChanging := nil;
  inherited Destroy;
  FStrArray.Free;
end;

function TJvRnnStrList.Find(const S: string; var Index: Integer): Boolean;
begin
  Result := FStrArray.GetItemIndex(Index, s);
end;

function TJvRnnStrList.IndexOf(const S: string): Integer;
begin
  if not Sorted then
    Result := inherited IndexOf(S) else
  if not Find(S, Result) then
    Result := -1;
end;

function TJvRnnStrList.Add(const S: string): Integer;
begin
  if not Sorted then
    Result := FStrArray.Count
  else
    if Find(S, Result) then
      case Duplicates of
        dupIgnore: Exit;
        dupError: Error(SDuplicateString, Result);
      end;
  InsertItem(Result, S);
end;

constructor TJvRnnStrList.Create;
begin
  inherited Create;
  FStrArray := TJvCCStrObjArray.Create(0);
  FStrArray.Sorted := FALSE;
end;

procedure TJvRnnStrList.Changed;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{$IFDEF VER80}
function TJvRnnStrList.GetStrName(ix:integer):string;
var
  ip : integer;
begin
  Result := JvRnnStrUt.GetStrName(FStrArray.ArrayPtr^[ix].sval^);
  Result := FStrArray.ArrayPtr^[ix].sval^;
  ip := Pos('=',Result);
  if ip > 0 then
    SetLength(Result,ip-1);
end;
{$ENDIF}

procedure TJvRnnStrList.Changing;
begin
  if Assigned(FOnChanging) then
    FOnChanging(Self);
end;

procedure TJvRnnStrList.Sort;
begin
  BeginUpdate;
  FStrArray.QuickSort;
  EndUpdate;
end;

procedure TJvRnnStrList.SetSorted(Value: Boolean);
begin
  if FSorted <> Value then
  begin
    if Value then Sort;
    FStrArray.Sorted := Value;
    FSorted := FStrArray.Sorted;
  end;
end;

procedure TJvRnnStrList.Clear;
begin
  if FStrArray.Count <> 0 then
  begin
    BeginUpdate;
    FStrArray.Count := -1;
    EndUpdate;
  end;
end;

procedure TJvRnnStrList.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FStrArray.Count) then Error(SListIndexError, Index);
  BeginUpdate;
  FStrArray.DeleteItemsAt(Index,1);
  EndUpdate;
end;

procedure TJvRnnStrList.PutObject(ix : integer; vobj : Tobject);
begin
  if (ix < 0) or (ix >= FStrArray.Count) then Error(SListIndexError, ix);
  BeginUpdate;
  FStrArray.ArrayPtr^[ix].oval := vobj;
  EndUpdate;
end;

function  TJvRnnStrList.GetObject(ix : integer) : TObject;
begin
  if (ix < 0) or (ix >= FStrArray.Count) then
    Error(SListIndexError, ix);
  Result := FStrArray.ArrayPtr^[ix].oval;
end;


function TJvRnnStrList.Get(Index: Integer): string;
begin
  if (Index < 0) or (Index >= FStrArray.Count) then
    Error(SListIndexError, Index);
{$IFDEF USE_PSTRING}
  Result := FStrArray.ArrayPtr^[Index].sval^;
{$ELSE}
  Result := FStrArray.ArrayPtr^[Index].sval;
{$ENDIF}
end;


function TJvRnnStrList.GetCount: Integer;
begin
  Result := FStrArray.Count;
end;


procedure TJvRnnStrList.Insert(Index: Integer; const S: string);
begin
  if Sorted then Error(SSortedListError, 0);
  InsertItem(Index, S);
end;

procedure TJvRnnStrList.InsertItem(Index: Integer; const S: string);
begin
  if (Index < 0) or (Index > FStrArray.Count) then Error(SListIndexError, Index);
  BeginUpdate;
  FStrArray.CreateItemsAt(Index,1);
{$IFDEF USE_PSTRING}
  ASSIGNSTR(FStrArray.ArrayPtr^[Index].sval,s);
{$ELSE}
  FStrArray.ArrayPtr^[Index].sval := s;
{$ENDIF}
  EndUpdate;
end;

procedure TJvRnnStrList.Put(Index: Integer; const S: string);
begin
  if Sorted then Error(SSortedListError, 0);
  if (Index < 0) or (Index >= FStrArray.Count) then Error(SListIndexError, Index);
  BeginUpdate;
  FStrArray.SetStr(Index,s);
  EndUpdate;
end;

{$IFNDEF VER80}
function TJvRnnStrList.GetCapacity: Integer;
begin
  Result := FStrArray.Capacity;
end;

procedure TJvRnnStrList.SetCapacity(NewCapacity: Integer);
begin
  FStrArray.Capacity := NewCapacity;
end;
{$ENDIF}

procedure TJvRnnStrList.SetUpdateState(Updating: Boolean);
begin
  if Updating then
    Changing
  else
    Changed;
end;

end.
 
