{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnStrUt.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnStrUt;


interface
uses SysUtils,classes;


{CIPOS: A Case insensitive POS routine}
function CIPos(const substr, s : string) : integer;

function StrToPChar(const Str: string): PChar;
function StrArrayToCStr(const inArray : array of string):PChar;
function StrListToCStr(slist : TStrings; separator:Char):PChar;

function StrListToString(slist : TStrings; separator:Char) : string;

procedure SplitNameValue(const s : string; var name, value:string);
procedure RemoveEndChar(var str:string; endChar:Char);
procedure EnsureEndChar(var str:string; endChar:Char);
function  GetParentDirectory(const inPath:string):string;
function  AssembleStr(const instrarray : array of string) : string;

function QuoteStr(const instr : string) : string;
function GetSQLDateStr(const inDate : TDateTime) : string;
function GetSQLTimeStr(const inDate : TDateTime) : string;
function GetSQLDateTimeStr(const inDate : TDateTime) : string;

procedure ReplaceStringsText(slist : TStrings; const token, replacement : string);
function  ReplaceTokenInString(var s : string; const token, replacement : string) : integer;
procedure ExtractFileNameExt(const inName:string; var outname, outext:string);
function RnnExtractPath(const inName:string) : string;
function ExtractBeforeDot(const s : string) : string;

{$IFNDEF WIN32}
function ExtractFileDir(const inPath : string) : string;
{$ENDIF}

procedure RnnStrPCopy(p : PChar; const s : string);

procedure EncryptStrings(const Key : string; slist : TStrings; dohex:boolean);
procedure DecryptStrings(const Key : string; slist : TStrings; unhex:boolean);

procedure HexToChar(const hex1,hex2 : Char; var outChar:Char);
procedure CharToHex(const inChar:Char; var hex1,hex2:Char);
function HexStrToInt(const instr : string) : integer;
procedure HexToBinary(phex : PChar; var memptr : Pointer; var memsize : integer);
function GetStrName(const instr : string):string;
function GetStrValue(const instr : string):string;

{$ifdef VER80}
type
  ShortString = string;
  PShortString = ^ShortString;
  AnsiChar = Char;
  PAnsiChar = ^AnsiChar;

procedure SetLength(var Str: string; Length: integer);
procedure SetString(var Str: string; From: PChar; Length: integer);
function Trim(const Str: string): string;
function TrimLeft(const Str: string): string;
function TrimRight(const Str: string): string;

{$endif}

implementation
type
  PByte = ^Byte;
{$IFNDEF WIN32}
type
  TRingIndex = 0..7;
var
  Ring: array[TRingIndex] of PChar;
  RingIndex: TRingIndex;

procedure ThrowStrLengthError;
begin
  raise EConvertError.create('String length exceeds 255');
end;



{$ENDIF}

procedure SplitNameValue(const s : string; var name, value:string);
var
  ichar : integer;
begin
  ichar := Pos('=',s);
  if (ichar > 0) then
  begin
    name := Copy(s,1,ichar-1);
    value := Copy(s,ichar+1,Length(s)-ichar);
  end
  else begin
    name := s;
    value := '';
  end;
end;


function RnnExtractPath(const inName:string) : string;
begin
  Result := ExtractFilePath(inName);
  RemoveEndChar(Result,'\');
end;

function ExtractBeforeDot(const s : string) : string;
var
  i :integer;
begin
  i := Pos('.',s);
  if (i > 1) then
    Result := Copy(s,1,i-1)
  else
    Result := s;
end;

{lifted from StrHlder, RxLib, procedurised}
function IsHex(hchar : char; var value : integer) : boolean;
begin
  if (hchar >= '0') and (hchar <= '9') then
  begin
    Result := TRUE;
    value := ORD(hchar) - ord('0');
  end else
  if (hchar >= 'A') and (hchar <= 'F') then
  begin
    Result := TRUE;
    value := ORD(hchar) - ORD('A') + 10;
  end else
  if (hchar >= 'a') and (hchar <= 'f') then
  begin
    Result := TRUE;
    value := ORD(hchar) - ORD('a') + 10;
  end else
  begin
    Result := FALSE;
  end;
end;

procedure HexToChar(const hex1,hex2 : Char; var outChar:Char);
var
  value1 : integer;
  value2 : integer;
begin
  if (not IsHex(hex1, value1)) or (not IsHex(hex2, value2)) then
    raise EConvertError.Create('Not a hexadecimal number');
  outChar := CHR((value1 SHL 4) + value2);
end;

procedure CharToHex(const inChar:Char; var hex1,hex2:Char);
const
  kHexNumbers : string = '0123456789abcdef';
var
  ix : integer;
begin
  ix := ORD(inChar) SHR  4;
  hex1 := kHexNumbers[ix+1];
  ix := ORD(inChar) and $000F;
  hex2 := kHexNumbers[ix+1];
end;

function UnhexaString(const Src : string) : string;
var
  i : integer;
  j : integer;
  k : integer;
begin
  k := Length(src);
  if (k mod 2) <> 0 then
    EConvertError.create('Invalid Hex String');
  SetLength(Result, k div 2);
  i := 1;
  j := 1;
  while (i < k) do
  begin
    HexToChar(Src[i],Src[i+1],Result[j]);
    Inc(i,2);
    Inc(j);
  end;
end;

function HexaString(const Src : string) : string;
var
  i : integer;
  j : integer;
  c1, c2 : Char;
begin
  SetLength(Result,Length(Src)*2);
  j := 1;
  for i := 1 to Length(Src) do
  begin
    CharToHex(Src[i],c1,c2);
    Result[j] := c1;
    Result[j+1] := c2;
    Inc(j,2);
  end;
end;

function XorString(const Key, Src : string; var keystart: integer) : string;
var
  I: Integer;
  keyix : integer;
  keylen : integer;
begin
  {xor and turn into hexadecimal}
  Result := Src;
  if Length(Key) > 0 then
  begin
    keyix := keystart;
    keylen := Length(Key);
    for I := 1 to Length(Src) do begin
      if (keyix < 1) or (keyix > keylen) then
        keyix := 1;
      Result[I] := Chr(Byte(Key[keyix]) xor Ord(Src[I]));
      Inc(keyix);
    end;
    keystart := keyix;
  end;
end;

procedure HexToBinary(phex : PChar; var memptr : Pointer; var memsize : integer);
{take a string of hex characters (must be even number),
 count them, assign memory for equivalent binary value.
 if memptr is nil and memsize 0, count them first and allocate memory
 else expect memptr points to memsize number of bytes.
 Ignores non-hexadecimal separators between pairs.
 }
var
  hexct : integer;

  procedure ThrowInvalidString;
  begin
    raise EConvertError.create('HexToBinary: Invalid hex string');
  end;

  function CountHex(p : PChar) : integer;
  var
    highvalue : integer;
    lowvalue : integer;
  begin
    Result := 0;
    while (p^ <> #0) do
    begin
      if IsHex(p^,highvalue) then
      begin {next must be hex pair}
        Inc(p);
        if not IsHex(p^,lowvalue) then
          ThrowInvalidString;
        Inc(p);
        Inc(Result);
      end else
        Inc(p);
    end;
  end;

  procedure CalculateHex(p : PChar);
  var
    highvalue : integer;
    lowvalue : integer;
    pcount : integer;
    pvalue : PByte;
  begin
    {memptr must not be nil, only read memsize worth of hex}
    pvalue := PByte(memptr);
    pcount := memsize;
    while (p^ <> #0) do
    begin
      if IsHex(p^,highvalue) then
      begin {next must be hex pair}
        Inc(p);
        if not IsHex(p^,lowvalue) then
          ThrowInvalidString;
        Inc(p);
        pvalue^ := (highvalue SHL 4) + lowvalue;  {in a byte, non-endian}
        Inc(pvalue);
        Dec(pcount);
        if (pcount = 0) then
          exit;
      end else
      begin
        {keep looking}
        Inc(p);
      end;
    end;
    ThrowInvalidString;
  end;

begin
  if (memsize > 0) then
    CalculateHex(phex)
  else begin
    hexct := CountHex(phex);
    if (hexct > 0) then
    begin
      GetMem(memptr,hexct);
      memsize := hexct;
      CalculateHex(phex);
    end;
  end;
end;

function HexStrToInt(const instr : string) : integer;
var
  mp : PByte;
  sz : integer;
  i : integer;
  p : PChar;
  vi : integer;
  vp : Pointer;
begin
  sz := sizeof(integer);
  mp := @Result;
  inc(mp,sz);  {prepare for backwards endian}
  p := StrToPChar(instr);
  vi := 1;
  for i := 0 to sz-1 do
  begin
    Dec(mp);
    vp := mp;
    HexToBinary(p, vp, vi);
    Inc(p,2);
  end;
end;

procedure DecryptStrings(const Key : string; slist : TStrings; unhex:boolean);
var
  i : integer;
  keyix : integer;
  s : string;
begin
  slist.BeginUpdate;
  try
    keyix := 1;
    for i := 0 to slist.Count-1 do
    begin
      s := slist[i];
      if unhex then
        s := UnhexaString(s);
      slist[i] := XorString(Key,s,keyix);
    end;
  finally
    slist.EndUpdate;
  end;
end;

procedure EncryptStrings(const Key : string; slist : TStrings; dohex:boolean);
var
  i : integer;
  keyix : integer;
  s : string;
begin
  slist.BeginUpdate;
  try
    keyix := 1;
    for i := 0 to slist.Count-1 do
    begin
      s := XorString(Key,slist[i],keyix);
      if dohex then
        s := HexaString(s);
      slist[i] := s;
    end;
  finally
    slist.EndUpdate;
  end;
end;

procedure ExtractFileNameExt(const inName:string; var outname, outext:string);
var
	 i : integer;
	 prefix : string;
  exlen : integer;
begin
	 prefix := ExtractFileName(inName);
	 for i := Length(prefix) downto 1 do
  begin
		  if prefix[i] = '.' then
    begin
      exlen := Length(prefix)-i;
      outname := Copy(prefix, 1, i-1);
      outext := Copy(prefix, i+1, exlen);
      exit;
		  end;
	 end;
  outname := prefix;
  outext := '';
end;

function CIPos(const substr, s : string) : integer;
var
  sslen : integer;
  slen : integer;
  si : integer;
  i : integer;
  schar : Char;
  sschar : Char;
  found : boolean;
begin
  Result := 0;
  sslen := Length(substr);
  slen := Length(s);

  if sslen > slen then begin
    exit;
  end;

  for  si := 0 to slen - sslen do
  begin
    found := true;
    for i := 1 to sslen do
    begin
      schar := s[si+i];
      sschar := substr[i];
      if (schar >= 'a') and (schar  <= 'z') then
        schar := CHAR( ORD(schar)-ORD('a') + ORD('A'));
      if (sschar >= 'a') and (sschar  <= 'z') then
        sschar := CHAR( ORD(sschar)-ORD('a') + ORD('A'));
      if sschar <> schar then
      BEGIN
        found := false;
        BREAK;
      END;
    end;
    if found then
    begin
      Result := si + 1;
      exit;
    end;
  end;
end;

function QuoteStr(const inStr : string) : string;
begin
 	Result := ''''+inStr+'''';
end;

function InternalDateTimeStr(const dsform:string; indt:TDateTime):string;
var
	s : string;
begin
  s := FormatDateTime(dsform,indt);
  Result := QuoteStr(s);
end;

function GetSQLDateStr(const inDate : TDateTime) : string;
begin
  Result := InternalDateTimeStr('dd-mmm-yyyy',inDate);
end;

function GetSQLDateTimeStr(const inDate : TDateTime) : string;
begin
  Result := InternalDateTimeStr('dd-mmm-yyyy hh:nn:ss',inDate);
end;

function GetSQLTimeStr(const inDate : TDateTime) : string;
begin
  Result := InternalDateTimeStr('hh:nn:ss',inDate);
end;

function  AssembleStr(const instrarray : array of string) : string;
var
	i : integer;
begin
	Result := '';
	for i := Low(instrarray) to High(instrarray) do begin
    	if ( i > Low(instrarray) ) then
        	Result := Result  + ' ';
        Result := Result + instrarray[i];
    end;
end;


function  ReplaceTokenInString(var s : string; const token, replacement : string) : integer;
var
    ip : integer;
    nextvalid : integer;
begin
  	ip := Pos(token,s);
    Result := 0;
    while (ip > 0) do begin
    	Inc(Result);
    	nextvalid := ip + Length(token);
        s := Copy(s,1,ip-1)+replacement+Copy(s,nextvalid,Length(s)-nextvalid+1);
        ip := Pos(token,s);
    end;
end;

procedure  ReplaceStringsText(slist : TStrings; const token, replacement : string);
var
  s : string;
  i : integer;
begin
  if Length(token)=0 then
    exit;
  for i := 0 to slist.count-1 do
  begin
    s := slist[i];
    if (ReplaceTokenInString(s,token,replacement) > 0) then
    		slist[i] := s;
  end;
end;


function StrListToString(slist : TStrings; separator:Char):string;
var
  i : integer;
  totalLength : integer;
  pc : PChar;
  slistCount : integer;
begin
  slistCount := slist.Count;
  if separator <> #0 then
    totalLength := slistCount-1
  else
    totalLength := 0;
  for i := 0 to slistCount-1 do
    totalLength := totalLength + Length(slist[i]);
  if (totalLength < 0) then
     totalLength := 0;
  SetLength(Result,totalLength);
  pc := StrToPChar(Result);
  for i := 0 to slistCount-1 do
  begin
    if (i > 0) and (separator <> #0) then
    begin
      pc^ := separator;
      Inc(pc);
    end;
    RnnStrPCopy(pc, slist[i]);
    Inc(pc,Length(slist[i]));
  end;
end;


function StrListToCStr(slist : TStrings; separator:Char):PChar;
var
  i : integer;
  totalLength : integer;
  pc : PChar;
  slistCount : integer;
begin
  slistCount := slist.Count;
  if separator <> #0 then
    totalLength := slistCount-1
  else
    totalLength := 0;
  for i := 0 to slistCount-1 do
    totalLength := totalLength + Length(slist[i]);
  if (totalLength < 0) then
     totalLength := 0;
  Result := StrAlloc(totalLength+1);
  pc := Result;
  for i := 0 to slistCount-1 do
  begin
    if (i > 0) and (separator <> #0) then
    begin
      pc^ := separator;
      Inc(pc);
    end;
    RnnStrPCopy(pc, slist[i]);
    Inc(pc,Length(slist[i]));
  end;
end;


function StrArrayToCStr(const inArray : array of string):PChar;
var
  i : integer;
  totalLength : integer;
  pc : PChar;
begin

  totalLength := High(inArray) - Low(inArray); {spaces}
  for i := Low(inArray) to High(inArray) do totalLength := totalLength
                                                         + Length(inArray[i]);
  if (totalLength < 0) then
     totalLength := 0;

  Result := StrAlloc(totalLength+1);
  pc := Result;
  for i := Low(inArray) to High(inArray) do
  begin
    if i > Low(inArray) then
    begin
      pc^ := ' ';
      Inc(pc);
    end;
    RnnStrPCopy(pc, inArray[i]);
    Inc(pc,Length(inArray[i]));
  end;

end;

procedure EnsureEndChar(var str:string; endChar:Char);
var
  ilen : integer;
begin
  ilen := Length(str);
  if (ilen > 0) and (str[ilen]<>endChar) then
  begin
    Inc(ilen);
    SetLength(str,ilen);
    str[ilen] := endChar;
  end;
end;

procedure RemoveEndChar(var str:string; endChar:Char);
var
	ilen : integer;
begin
	ilen := Length(str);
	if (ilen > 0) and (str[ilen]=endChar) then
		SetLength(str,ilen-1);
end;
{$IFNDEF WIN32}
function ExtractFileDir(const inPath : string) : string;
begin
  Result := ExtractFilePath(inPath);
  RemoveEndChar(Result,'\');
end;
{$ENDIF}

function GetParentDirectory(const inPath:string):string;
var
	i : integer;
begin
    Result := '';

    i := Length(inPath);
    if (i = 0) then
    	exit;

    if inPath[i]='\' then
    	Dec(i);
    while ( i >= 1) do begin
        if (inPath[i]='\') then begin
            Result := Copy(inPath,1,i-1);
            exit;
        end;
        Dec(i);
    end;
end;

function GetStrName(const instr : string):string;
var
  i : integer;
begin
  i := Pos('=',instr);
  if (i > 0) then
    Result := Copy(instr,1,i-1)
  else
    Result := instr;
end;

function GetStrValue(const instr : string):string;
var
  i : integer;
begin
  i := Pos('=',instr);
  if (i > 0) then
    Result := Copy(instr,i+1,Length(instr)-i)
  else
    Result := instr;
end;

{not multi-byte}
procedure RnnStrPCopy(p : PChar; const s : string);
var
  i : integer;
begin
  {$ifdef VER80}
  StrPCopy(p, s);
  {$else}
  for i := 1 to Length(s) do
  begin
    p^ := s[i];
    Inc(p);
  end;
  p^ := #0;
  {$endif}
end;

{$ifndef VER80}

{
  Delphi 2.0 knows how to convert string to PChar.
}
function StrToPChar(const Str: string): PChar;
begin
  Result := PChar(Str);
end;

{$else}


{
  Return a PChar representation of the string, Str. Allocate a dynamic
  copy of the string. Keep a ring of 8 dynamic strings, and free the
  old strings. Thus, you can usually rely on the returned string being
  valid while it is needed. The most common need is to pass an argument
  to a Windows API function, so the need is temporary, but several
  such strings might be required. That's why the ring has 8 items in it:
  more than enough for most uses.
}


function StrToPChar(const Str: string): PChar;
begin
  { Allocate a PChar and copy the original string. }
  Result := StrAlloc(Length(Str)+1);
  RnnStrPCopy(Result, Str);

  { Add the string to the ring. }
  StrDispose(Ring[RingIndex]);
  Ring[RingIndex] := Result;
  RingIndex := (RingIndex + 1) mod (High(TRingIndex) + 1);
end;

{
  Set the length of a string.
}
procedure SetLength(var Str: string; Length: integer);
begin
  if Length > 255 then
    ThrowStrLengthError;
  Str[0] := Chr(Length)
end;

{
  Set the contents of a string. If there are fewer than Length bytes
  in the string, From, then leave the remaining bytes unchanged.
}
procedure SetString(var Str: string; From: PChar; Length: integer);
var
  FromLen: Integer;
begin
  if From <> nil then
  begin
    { Only copy as many bytes as are in the From string. }
    FromLen := StrLen(From);
    if FromLen < Length then
      Length := FromLen;
    SetLength(Str,Length);
    Move(From^, Str[1], Length);
  end;
end;

{
  Return whether the character C, is a white space character,
  or a nonprintable control character.
}
function IsWhiteSpace(C: Char): Boolean;
begin
  Result := C in [#0..' ']
end;

{
  Trim all leading and trailing white space characters.
}
function Trim(const Str: string): string;
var
  L, R: Integer;
begin
  L := 1;
  R := Length(Str);
  while (L < R) and IsWhiteSpace(Str[L]) do
    Inc(L);
  while (L < R) and IsWhiteSpace(Str[R]) do
    Dec(R);
  Result := Copy(Str, L, R-L+1);
end;

{
  Trim leading white space characters.
}
function TrimLeft(const Str: string): string;
var
  L, R: Integer;
begin
  L := 1;
  R := Length(Str);
  while (L < R) and IsWhiteSpace(Str[L]) do
    Inc(L);
  Result := Copy(Str, L, 255);
end;

{
  Trim trailing white space characters.
}
function TrimRight(const Str: string): string;
var
  R: Integer;
begin
  R := Length(Str);
  while (1 < R) and IsWhiteSpace(Str[R]) do
    Dec(R);
  Result := Copy(Str, 1, R);
end;


{
  Free all the left over strings in the StrToPChar ring.
}

procedure Terminate; far;
var
  I: TRingIndex;
begin
  for I := Low(TRingIndex) to High(TRingIndex) do
  begin
    StrDispose(Ring[I]);
    Ring[I] := nil; { just in case StrToPChar is called again }
  end;
end;

initialization
  AddExitProc(Terminate);
{$endif}
end.

