{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnStrms.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnStrms;

interface
Uses classes, sysutils;
type
     TEOLString = string[3];

     EStreamLineReader = class(Exception);

     PByte = ^Byte;

     TJvRnnBufStream = class(TStream)
     private
       FStream : TStream;

        FBuf : PChar;
        FBufSize : integer;
        FBufPos : integer;
    FBufChars: integer;  {valid characters in buffer}

        FStreamSize : LongInt;  {cached if read-only}
        FStreamPos : LongInt;   {position in stream of buffer start}
        FBufferDirty : boolean; {caused by a write}
        FOwnsStream : boolean;
        FFixedSize : boolean; {allows optimized look-up for size}

        function InternalBufRead(var buf; ct : LongInt):LongInt;
        function InternalBufWrite(const buf; ct : LongInt):LongInt;

        function AdvanceBuffer : LongInt;
    function FlushBuffer : LongInt;

        {optimised? raw stream functions}
        function GetStreamPos:LongInt;
        function GetStreamSize:LongInt;
        function GetReadCharsAvail:LongInt;

        procedure SetStreamPos(pos:LongInt);

     public
        constructor create(bstream : TStream; bufsize:integer);

        destructor  destroy; override;

        function PosFromEnd : LongInt;

      function Read(var Buffer; Count: LongInt): LongInt; override;
      function Write(const Buffer; Count: LongInt): LongInt; override;
      function Seek(Offset: LongInt; Origin: Word): LongInt; override;
        procedure SetFixedSize(fixsize:boolean);

        property FixedSize : boolean read FFixedSize write SetFixedSize;
     end;

     TJvRnnTextStream = class(TJvRnnBufStream)

       constructor createfromfile(const fname:string;Mode: Word);

     end;

   TJvRnnStreamRead =  class(TObject)
     private
       FStreamBuffer : TJvRnnBufStream;  {stream we are reading from}

        FLineBuffer : Pointer; {current line accumulates here}
     FMaxLineSize : Cardinal;  {size of line buffer}
    FLinePos : Cardinal;

        dynlinesize : Boolean; {if maxline was zero, allow line buffer to increase to largest line}
    procedure SetLineBufSize(inSize:integer);
    procedure SetSizeIsfixed(sizefixed:boolean);
        function GetSizeIsFixed : boolean;

     public
        constructor create(inStream:TStream; maxline:Cardinal);
        destructor  destroy; override;

        function NextLine : PChar;  {returns pointer to next lines text, nil if eof}
        function LineLength : integer; {current lines length}

        property FixedSize : boolean read GetSizeIsFixed write SetSizeIsFixed;
     end;



implementation
{-------------------------------------}
Uses JvCCArray;

constructor TJvRnnTextStream.createfromfile(const fname:string;Mode: Word);
var
  aStream : TStream;
begin
  aStream := TFileStream.Create(fname, Mode);
    FOwnsStream := TRUE;
  inherited create(aStream,8192);
end;


constructor TJvRnnBufStream.create(bstream : TStream; bufsize:integer);
begin
  inherited create;
  FStream := bstream;
    FBufSize := bufsize;
    GetMem(FBuf,FBufSize);
end;


destructor  TJvRnnBufStream.destroy;
begin
  if FBufferDirty then
      FlushBuffer;

  FreeMem(FBuf,FBufSize);
    if FOwnsStream then
      FStream.Free;
    inherited destroy;
end;


function TJvRnnBufStream.PosFromEnd : LongInt;
begin
   Result := GetReadCharsAvail - FBufPos;
end;


procedure TJvRnnBufStream.SetFixedSize(fixsize:boolean);
begin
  if fixsize then
      FStreamSize := FStream.Size;{remember size}
    FFixedSize := fixsize;
end;


function TJvRnnBufStream.GetReadCharsAvail:LongInt;
begin
  if FFixedSize then begin
      Result := FStreamSize - FStreamPos;
    end else begin
    Result := FStream.Seek(0, soFromEnd); {go end and get offset}
      Result := Result - FStream.Seek(FStreamPos, soFromBeginning); {subtract and go to current position}
    end;
end;

function  TJvRnnBufStream.GetStreamPos:LongInt;
begin
  Result := FStreamPos + FBufPos;
end;

function  TJvRnnBufStream.GetStreamSize:LongInt;
begin
  Result := FStream.Size;
end;

function TJvRnnBufStream.Seek(Offset: LongInt; Origin: Word): LongInt;
var
  oldPosition : LongInt;
begin
  oldPosition := GetStreamPos;

  case Origin of
  soFromBeginning: Result := Offset;
  soFromCurrent:   Result := GetStreamPos + Offset;
  soFromEnd:       Result := GetStreamSize - Offset;
  else
    raise Exception.Create('Invalid stream origin '+IntToStr(Origin));
  end;

  if Result <> oldPosition then
  begin
    SetStreamPos(Result);
  end;
end;


function TJvRnnBufStream.Write(const Buffer; Count: LongInt): LongInt;
var
  availChars : LongInt;
  outct : LongInt;
  outPtr : PByte;
begin
  if (FBufChars > 0) then
    SetStreamPos(GetStreamPos);
  Result := Count;
  outct := Count;
  outPtr := @Buffer;
  while (outct > 0) do
  begin
    availChars := FBufSize - FBufChars; {buffer space not yet written to}
    if (availChars <= 0) then
      FlushBuffer;
    availChars := InternalBufWrite(outPtr^, outct);
    if (availChars <= 0) then
    begin
      FlushBuffer;
    end else
    begin
      Dec(outct,availChars);
      Inc(outPtr,availChars);
    end;
  end;
end;

function TJvRnnBufStream.FlushBuffer : LongInt;
var
  charsWrit : LongInt;
begin    {make room in buffer for more input}
  if (FBufPos > 0) then begin
      charsWrit := FStream.Write(FBuf^, FBufPos);
      Inc(FStreamPos,charsWrit);
      FBufPos := 0;
    end;
    FBufferDirty := FALSE;
    Result := FBufSize;
end;


function TJvRnnBufStream.AdvanceBuffer : LongInt;
var
  availChars : LongInt;
begin
  Result := FBufSize;
    availChars := FStream.Size - FStreamPos;
    if Result > availChars then
      Result := availChars;
    Inc(FStreamPos,FBufChars); {point to current position}
    Result := FStream.Read(FBuf^, Result);
    FBufChars := Result;
    FBufPos := 0;
end;


procedure  TJvRnnBufStream.SetStreamPos(pos:LongInt);
begin

  if FBufferDirty then
      FlushBuffer; {no valid write buffer characters}

    FBufChars := 0; {no valid read buffer characters}

    FStream.Position := pos; { Reset stream. }
    FStreamPos := pos;
    FBufPos := 0;

end;

function  TJvRnnBufStream.InternalBufWrite(const buf; ct : LongInt):LongInt;
var
  charsAvail : LongInt;
  procedure CallFlush;
  begin
    charsAvail := FBufSize - FBufPos;
    if charsAvail = 0 then
      FlushBuffer;
    charsAvail := FBufSize - FBufPos;
  end;
begin
  CallFlush;
  if ct > charsAvail then
    ct := charsAvail;
  System.Move(buf, FBuf[FBufPos], ct);
  Inc(FBufPos,ct);
  Result := ct;
  FBufferDirty := TRUE;
  CallFlush;
end;


function TJvRnnBufStream.InternalBufRead(var buf; ct : LongInt):LongInt;
var
  charsAvail : LongInt;
begin
  {read some characters in the buffer only}
  charsAvail := FBufChars - FBufPos;
    if ct > charsAvail then
      ct := charsAvail;
    System.Move(FBuf[FBufPos],buf, ct);
    Inc(FBufPos,ct);
    Result := ct;
end;

function TJvRnnBufStream.Read(var Buffer; Count: LongInt): LongInt;
var
  charsAvail : LongInt;
    charsForward : LongInt;
    outPtr : PByte;
begin
  {read rct characters into buf}
    if FBufferDirty then
      FlushBuffer;

    charsAvail := PosFromEnd;
    if (Count > charsAvail) then
      Count := charsAvail;
    Result := Count;
    outPtr := @Buffer;
    while (Count > 0) do begin
    charsAvail := FBufChars - FBufPos;
        if (charsAvail <= 0) then
          AdvanceBuffer;

        charsForward := InternalBufRead(outPtr^,Count);
    Inc(outPtr,charsForward);
        Dec(Count,charsForward);
    end;
end;


constructor TJvRnnStreamRead.create(inStream:TStream; maxline:Cardinal);
begin
  FStreamBuffer := TJvRnnBufStream.Create(inStream,4096);

    dynlinesize := (maxline=0);

    if dynlinesize then
      maxline := 256;

    SetLineBufSize(maxline);

end;

destructor  TJvRnnStreamRead.destroy;
begin
    SetLinebufSize(0);
    FStreamBuffer.Free;
    inherited destroy;

end;

procedure TJvRnnStreamRead.SetLineBufSize(inSize:integer);
begin
{$IFDEF WIN32}
  ReallocMem(FLineBuffer,insize);
{$ELSE}
  ReallocMem16(FLineBuffer,insize);
{$ENDIF}
  FMaxLineSize := inSize;
end;


{NextLine: returns pointer to next lines text, nil if eof}

procedure TJvRnnStreamRead.SetSizeIsfixed(sizefixed:boolean);
begin
  FStreamBuffer.FixedSize := sizefixed;
end;

function TJvRnnStreamRead.GetSizeIsFixed : boolean; 
begin
  Result := FStreamBuffer.FixedSize;
end;


function  TJvRnnStreamRead.NextLine : PChar;
var
    nextchar : Char;
    charsLeft : LongInt;
    newSize : Cardinal;
    line : PChar;
begin
  Result := nil;

    FLinePos := 0;
    line := PChar(FLineBuffer);

  WITH FStreamBuffer do begin
      charsLeft := PosFromEnd;
        if charsLeft <= 0 then
          exit;

    while (charsLeft > 0) do begin
          Read(nextChar,1);
            Dec(charsLeft);

            if (nextChar=#10) then
        break;

      if (FLinePos >= FMaxLineSize-1) then begin
              if not dynlinesize then
          raise EStreamLineReader.Create('Line read buffer full');

                if FMaxLineSize < 4096 then
                    newSize := ((FMaxLineSize div 512) + 1)*512
                else
                  newSize := FMaxLineSize + 4096;

              if (newSize > 32000) then
          raise EStreamLineReader.Create('Line exceeds 32000');
        SetLineBufSize(newSize);
        line := PChar(FLineBuffer);

      end;


      line[FLinePos] := nextChar;
      Inc(FLinePos);
        end; {while}
    end;{with FStream}
    if (FLinePos > 0) and (line[FLinePos-1]=#13) then
      Dec(FLinePos);{ don't want terminator in line }
    line[FLinePos] := #0;
    Result := line;
end;

function TJvRnnStreamRead.LineLength : integer; {current lines length}
begin
  Result := FLinePos;
end;

end.
