{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnUtBDE.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnUtBDE;

interface

uses DB, DbiTypes,DBTables, Classes, SysUtils;
	 { user-created application units }
type
  TQ_Field = string[31];
  TQ_Type = string[31];

  TQ_FieldArray = array[0..0] of TQ_Field;
  TQ_TypeArray = array[0..0] of TQ_Type;

type
  EDBToolsException = class(Exception);
{$IFDEF WIN32}
type
	TRyDataset = TBDEDataset;
  int32 = integer;
{$ELSE}
type
	TRyDataset = TDataset;
  DbiDate = Longint;
  Time = Longint;
  int32 = Longint;
{$ENDIF}
procedure DBPackTable(inDB : TDatabase; inTable : TTable);

{Function CreateAnswerTable(inDB : TDatabase; const TblName :string; SQL: PChar): Longint;}
function GetSQLFieldDef(fieldDef : TFielddef) : string;

function FileNameMinusExtension(const inStr : string):string;
procedure OpenDatabasePath(inDb:TDatabase;const inPath : string);

procedure CreateFieldsFromDefs(inTable:TDataset);
procedure MakePrimaryIndex(Tbl : TTable; keyct : integer);

procedure DisposeIndexName(inTable : TTable; const inName:string);
function TableHasIndexName(inTable : TTable; const inName:string):Boolean;

procedure ExecuteCreateTableSQL(inDB: TDatabase;
                                      const intablename : string;
									  const InArray: array of string);

procedure SetListStrArray(slist: TStrings; const inArray : array of string);
procedure ExecuteQuerySQL(inQuery: TQuery; const InArray: array of string);


procedure ExecuteDatabaseSQL(inDB: TDatabase; const sql : string);
procedure ExecuteAliasSQL(const aliasDB: string; const sql : string);

procedure ExecuteBDESQL(hDB: HDBIDB; const sql : string);
procedure ExecuteBDESQL_StrList(hDB: HDBIDB; slist : TStrings);

function  DBTableRowCount(inDB : TDatabase; const tableName : string): int32;

function  DBOpenTable(inDB:TDatabase; const tableName: string) : TTable;
function DBOpenQuery(inDB : TDatabase; const sql : string): TQuery;

function  DBTableExists(inDB: TDatabase; const tableName: string): boolean;
procedure DeleteDBTable(inDB: TDatabase; const tableName: string);
procedure RenameTTable(Table: TTable; NewName: String; ReOpen: Boolean);

function ApplyDatasetToTable(
  inQuery: TRyDataset; const inTable: TTable;
  fieldListCopy: TStrings; bmode : TBatchMode; emptytable : boolean) : integer;

function QueryAsTable(inQuery: TQuery; const inTableName: string;
		      fieldListCopy: TStrings; emptytable : boolean): TTable;

procedure CreateTableSQL(const alias, tableName: string;
                         const fields: array of TQ_Field;
			 const fieldtypes: array of TQ_Type;
			 numfields, numPrimary: integer);

procedure PackParadoxTable(tabdb : TDatabase; const inTabName : string);
procedure DBEmptyTable(inDB : TDatabase; const tableName : string);

function StrArrayToStr(const inSQL: array of string): string;
procedure DeleteTableSQL(const alias, tablename: string);
function QuoteCommaList(slist: TStrings): string;
function  EncloseInQuotes(const inStr : string) : string;

procedure GetFieldList(db:TDatabase; const tablename : string;
  slist : TStrings);

Function GetTableLevel(inDB : TDatabase; const inTabName : string) : WORD;
procedure DBRenameTable(inDB : TDatabase; const oldName, newName : string);
Procedure UpgradeLevel(inDB : TDatabase; const inName : string);

procedure  DBAddIndexToTable(inDB:TDatabase; const intable, ixName, fieldList:string;
								options : TIndexOptions);

function CloneTTable(srcTable : TTable; destDB : TDatabase;
					const destname: string) : TTable;

Function InstantiateQuery(inQuery : TQuery;
							const inTableName : string;
							keyCount : integer) : TTable;

Function InstantiateQueryDB(inQuery : TQuery;
  destDB : TDatabase;
  const inTableName : string;
  keyCount : integer) : TTable;

Function InstantiateAnswerStr(srcDB, destDB : TDatabase;
							sql:PChar;
                            inTableName:string;
							keyCount : integer):TTable;

procedure FillStringsFromQuery(
  inDB : TDatabase;
  const sql : string;
  slist : TStrings);

Function InstantiateAnswer(srcDB, destDB : TDatabase;
							sql : TStrings;
							const inTableName : string;
							keyCount : integer) : TTable;
{GetDBDirectory
Returns path to the databases directory. Database must be connected.
Strips final '\' off path.}

function GetDBDirectory(inDB : TDatabase) : string;

Function GetDatabaseFromAlias(const inAlias:string):TDatabase;

function GetDBDriverName(db:TDatabase):string;

procedure BatchCopyTable(const tabName : string;
						srcdb,destdb : TDatabase);

{GetDBDirectory;
Ensure TDatabase is open and return its Handle.
}
function DelphiDateTimeToBDE(inDateTime:TDateTime) : double;
function BDEToDelphiDateTime(inDate:double) : TDateTime;

function GetConnectedDBHandle(inDB : TDatabase) : hDBIDB;


function OracleTimeStampStr(inv : TDateTime) : string;
function SQLDateStr(inv : TDateTime) : string;
function SQLDateStrQuote(inv : TDateTime) : string;
function StdTimeStampStr(inv : TDateTime) : string;
function StdTimeStampQuote(inv : TDateTime) : string;

implementation
uses DbiProcs, DbiErrs, Forms, JvRnnStrUt;


{$IFDEF VER100}
procedure FreeAndNil(var obj:TObject);
var
  saveobj : TObject;
begin
  saveobj := obj;
  obj := nil;
  saveobj.free;
end;
{$ENDIF}

function OracleTimeStampStr(inv : TDateTime) : string;
begin {not really oracle specific}
{formatted to match TO_DATE(result,'DD-MON-YYYY HH24:MI:SS')}
  Result := FormatDateTime('dd-mmm-yyyy hh:nn:ss',inv);
end;

function SQLDateStr(inv : TDateTime) : string;
begin
  Result := FormatDateTime('dd-mmm-yyyy',inv);
end;

function SQLDateStrQuote(inv : TDateTime) : string;
begin
  Result := QuoteStr(SQLDateStr(inv));
end;

function StdTimeStampStr(inv : TDateTime) : string;
begin
  Result := FormatDateTime('mm/dd/yyyy hh:nn:ss',inv);
end;

function StdTimeStampQuote(inv : TDateTime) : string;
begin
  Result := QuoteStr(StdTimeStampStr(inv));
end;

function GetDBasePhysicalType(desc:FLDDesc) :string;
begin
  CASE desc.iFldType of
    fldDBCHAR : Result := Format('VARCHAR(%d)',[desc.iunits1]);
    fldDBNUM : Result := Format('NUMERIC(%d,%d)',[desc.iunits1,desc.iunits2]);
    fldDBMEMO : Result := Format('BLOB(%d,1)',[desc.iunits1]);
    fldDBBOOL : Result := 'BOOLEAN';
    fldDBDATE : Result := 'DATE';
    fldDBFLOAT : Result :=  Format('BLOB(%d,0)',[desc.iunits1]);
  ELSE
  	  raise EDatabaseError.Create('Unknown DBase physical type');
  END;
end;


procedure DBEmptyTable(inDB : TDatabase; const tableName : string);
begin
  Check(DbiEmptyTable (inDB.Handle,nil,StrToPChar(tableName),nil));
end;

procedure RenameTTable(Table: TTable; NewName: String; ReOpen: Boolean);
var
  hDb: hDBIDb;
  Props: CURProps;

begin
  if (not Table.Exclusive) then
  begin
    Table.Active := FALSE;
    Table.Exclusive := TRUE;
  end;
  Table.Active := TRUE;
  Check(DbiGetCursorProps(Table.Handle, Props));

  {Get the Database Handle from the table cursor since Table.DBHandle will
  be invalid once the table is closed}

  Check(DbiGetObjFromObj(hDBIObj(Table.Handle), objDATABASE, hDBIObj(hDb)));
  Table.Close;
  Check(DbiRenameTable(hDb, StrToPChar(Table.TableName), Props.szTableType, StrToPChar(NewName)));
  if ReOpen = True then
  begin
    Table.TableName := NewName;
    Table.Open;
  end;
end;


function OpenCursorNoXLate(inDB:TDatabase; const intabName:string ) : hDBICur;
begin
  Check(DbiOpenTable(inDB.Handle, StrToPChar(intabName), nil, nil, nil, 0,
    			                dbiREADONLY, dbiOPENSHARED,
                       xltNONE,  TRUE, nil, Result));
end;

procedure AppendSQLFieldDefs(inDB:TDatabase; const inName:string; outList : TStrings);
var
  curProp : CurProps;
  pfldDes  : pFLDDesc;
  pCurFld : pFLDDesc;
  i : integer;
  memsize : integer;
  tempStr : string;
  tabledrv : string;
  tabletype : TTableType;
  xcur:hDBICur;
begin
	 xcur := OpenCursorNoXLate(inDB, inName);
	 Check(DbiGetCursorProps(xcur,curProp));
  TRY
    tabledrv := StrPas(curProp.szTableType);

    if tabledrv=szDBASE then
      tabletype := ttDBase
    else if tabledrv= szPARADOX then
      tabletype := ttParadox
    else if tabledrv= szASCII then
      tabletype := ttASCII
    else
      tabletype := ttDEFAULT;

	   MemSize := curProp.iFields * Sizeof(FLDDesc);
	   pfldDes := AllocMem(MemSize);
    TRY
      pCurFld := pfldDes;
      Check( DbiGetFieldDescs(xcur, pfldDes) );
      i := 0;
      tempStr := '';
      while ( i < curProp.iFields) do
      begin
        if (i > 0) then
        begin
          tempStr := tempStr + ',';
          outList.Add(tempStr);
        end;
        tempStr := Format('%s ',[pCurFld^.szName]);

        CASE tabletype of
          ttDBASE : tempStr := tempStr + GetDBasePhysicalType(pCurFld^);
        END;

        Inc(pCurFld);
        Inc(i);
      end;
      outList.Add(tempStr);
    FINALLY
      FreeMem(pfldDes,MemSize);
    END;
  FINALLY
	   DbiCloseCursor(xcur);
  END;
end;

procedure AppendCreateTableSQL(inDB:TDatabase; const intabname:string;
									 outList : TStrings);
begin
  outList.Add('CREATE TABLE '''+ intabname + ''' (');
	 AppendSQLFieldDefs(inDB,intabname,outList);
	 outList.Add(');');
end;

procedure  DBAddIndexToTable(inDB:TDatabase;
			              const intable, ixName, fieldList:string;
                 options : TIndexOptions);
var
  rtab : TTable;
begin
  rtab := TTable.Create(nil);
  TRY
    rtab.Exclusive := TRUE;
    rtab.DatabaseName := inDB.Databasename;
    rtab.TableName := intable;
    rtab.IndexDefs.Update;

    if not TableHasIndexName(rtab,ixName) then
      rtab.AddIndex(ixName,fieldList,options);
  FINALLY
    rtab.Free;
  END;
end;

Function GetPrimaryIndexDef(inTable: TTable) : TIndexDef;
var
  i : integer;
begin
  inTable.IndexDefs.Update;
  for i := 0 to inTable.IndexDefs.Count-1 do
  begin
		  Result := inTable.IndexDefs[i];
		  if (ixPrimary in Result.Options) then
			   exit;
		end;
	 Result := nil;
end;

function GetConnectedDBHandle(inDB : TDatabase) : hDBIDB;
begin
  if not inDB.Connected then
    inDB.Connected := TRUE;
  Result := inDB.Handle;
end;

Function GetDatabaseFromAlias(const inAlias:string):TDatabase;
var
  MaxDbCount:integer;
  aliasList : TStringList;
begin
  with Session do
  begin {look at open databases first}
    if DatabaseCount > 0 then
    begin
      for MaxDbCount := 0 to DatabaseCount-1 do
      begin
        Result := Databases[MaxDbCount];
        if COMPARETEXT(inAlias,Result.DatabaseName)=0 then
          exit;
      end;
    end;
  end;

  (* not found, see if it is a BDE Alias *)

  aliasList := TStringList.Create;
  try
    Session.GetAliasNames(aliasList);
    if (aliasList.IndexOf(inAlias) < 0) then
    begin
      result := nil;
      exit;
    end;

  finally
    aliasList.free;
  end;
  {it is a BDE Alias, create a TDatabase for it
  Note that owner will be nil, and for proper
  cleanup, could be set by a call to InsertComponent
  for some component}

  Result := TDatabase.Create(nil);
  try
    Result.AliasName := inAlias;
    Result.DatabaseName := inAlias;
    Result.Connected := TRUE;
  except
    Result.Free;
    raise;
  end;
end;


procedure BatchCopyTable(const tabName : string;
						srcdb,destdb : TDatabase);
var
	srcCur : hDBICur;
	destDesc : BATTblDesc;

	probRecs, keyRecs, changeRecs, moveRecs : LongInt;
begin
   srcCur := OpenCursorNoXLate(srcdb, tabName);

   try
	 FillChar(destDesc,sizeof(BATTblDesc),#0);
	 destDesc.hDb	 := destdb.Handle;
	 StrPCopy(destdesc.szTblName, tabName);
	 moveRecs := 0;
	 probRecs := 0;
	 keyRecs := 0;
	 changeRecs := 0;

	 Check(DbiBatchMove(nil,srcCur, @destDesc,nil, batchCOPY,
						0, nil,nil,nil, 0, nil,nil,nil,
						@probRecs,@keyRecs, @changeRecs,
						false, false, moveRecs, false));


   finally
	 if srcCur <> nil then
		Check(DbiCloseCursor(srcCur));
   end;
end;



function CloneTableFields(srcTable:TTable) : TTable;
begin
	Result := nil;
	TRY
		Result := TTable.Create(nil);
		srcTable.FieldDefs.Update;
		Result.FieldDefs.Assign(srcTable.FieldDefs);
	EXCEPT
		Result.Free;
        raise;
	END;
end;

function CloneTableStruct(srcTable:TTable) : TTable;
var
	primaryKey : TIndexDef;
begin
	Result := nil;
	TRY
		Result := CloneTableFields(srcTable);
		primaryKey := GetPrimaryIndexDef(srcTable);
		if primaryKey <> nil then
			Result.IndexDefs.Add('',primaryKey.Fields, primaryKey.Options);
	EXCEPT
		Result.Free;
        raise;
	END;
end;

procedure MakePrimaryIndex(Tbl : TTable; keyct : integer);
var
  NewIndex: IDXDesc;
  i : integer;
begin

  if Tbl.Active then begin
	if Tbl.Exclusive = False then
		raise EDatabaseError.Create
			('TTable.Exclusive must be set to true in order to add an index to the table');
  end else begin
	Tbl.Exclusive := TRUE;
	Tbl.Active := TRUE;
  end;

  if GetPrimaryIndexDef(tbl)<>nil then
	raise EDatabaseError.Create('Table already has a primary key');

  NewIndex.iIndexId:= 0;
  NewIndex.bPrimary:= TRUE;
  NewIndex.bUnique:= TRUE;
  NewIndex.bDescending:= FALSE;
  NewIndex.bMaintained:= TRUE;
  NewIndex.bSubset:= FALSE;
  NewIndex.bExpIdx:= FALSE;
  NewIndex.iFldsInKey:= keyct;
  for i := 1 to keyct do
  	NewIndex.aiKeyFld[i-1]:= i;

  NewIndex.bCaseInsensitive:= FALSE;
  Check(DbiAddIndex(Tbl.dbhandle, Tbl.handle, StrToPChar(Tbl.TableName),
          szParadox, NewIndex, nil));
  Tbl.Close; {invalid table information now!}
  tbl.Open;
end;


(*Function CreateAnswerTable(inDB : TDatabase; const TblName :string; sql : PChar): Longint;
var
  hTmpDb: hDBIDB;
  hStmt: hDBIStmt;
  hQryCur, hNewCur: hDBICur;
  iRecCount: LongInt;

begin
  hQryCur := nil;
  hNewCur := nil;
  hStmt := nil;
  hTmpDb := inDB.Handle;
  try
{$IFDEF VER80}
	Check(DbiQPrepare(hTmpDb, qrylangSQL, sql, hStmt));
	Check(DbiQExec(hStmt, @hQryCur));

	Check(DbiQInstantiateAnswer(hStmt, hQryCur, StrToPChar(TblName), szPARADOX,
									  true, hNewCur));
{$ELSE}
	Check(DbiQAlloc(hTmpDb, qrylangSQL, hStmt));
	Check(DbiQPrepare(hStmt, sql));
	Check(DbiQExec(hStmt, @hQryCur));

	Check(DbiQInstantiateAnswer(hStmt, hQryCur, PChar(TblName), szPARADOX,
									  true, @hNewCur));
{$ENDIF}

	Check(DbiGetRecordCount(hNewCur, iRecCount));
	Result := iRecCount;
  finally

	if hStmt <> nil then
	  Check(DbiQFree(hStmt));
	if hNewCur <> nil then
	  Check(DbiCloseCursor(hNewCur));
  end;
end;*)

function  DBTableRowCount(inDB : TDatabase; const tableName : string): int32;
var
  hCur : hDBICur;
begin
  hCur := OpenCursorNoXLate(inDB,tablename);
   try
    Check(DbiGetRecordCount(hCur, Result));
   finally
    Check(DbiCloseCursor(hCur));
   end;
end;

Function InstantiateAnswerStr(srcDB, destDB : TDatabase;
							sql:PChar;
                            inTableName:string;
							keyCount : integer):TTable;
var
  qry : TQuery;

begin
  Result := nil;
  TRY
    qry := TQuery.create(nil);
    try
      qry.databasename := srcDB.databasename;
      qry.SQL.add(StrPas(sql));
      qry.open;

      Result := TTable.Create(nil);
      Result.databasename := destDB.databasename;
      Result.TableName := inTableName;
      Result.Exclusive := TRUE;
      Result.BatchMove(qry,batCopy);

    finally
      qry.free;
    end;

    if (keycount > 0) then begin
      MakePrimaryIndex(Result,keyCount);
    end;

  EXCEPT
    Result.Free;
    raise;
  END;
end;

Function InstantiateAnswer(srcDB, destDB : TDatabase;  sql : TStrings;
							const inTableName : string;
							keyCount : integer) : TTable;
var
	cstr : PChar;
begin
	cstr := StrListToCStr(sql,' ');
	try
	Result := InstantiateAnswerStr(srcDB,destDB, cstr, inTableName,keycount);
	finally
	StrDispose(cstr);
    end;
end;




function GetDBDirectory(inDB : TDatabase) : string;
var
	dirPath : PChar;
    slen : integer;
    hDB : hDBIDB;
begin
	hDB := GetConnectedDBHandle(inDB);
	dirPath := StrAlloc(DBIMAXPATHLEN+1);
	Result := '';
	TRY
		Check(DbiGetDirectory(hDB, FALSE, dirPath));
		Result := StrPas(dirPath);
        slen := Length(Result);
        if (slen > 0)  and (Result[slen] ='\') then
        	SetLength(Result,slen-1);
	FINALLY
		StrDispose(dirPath);
	END;
end;

procedure GetFieldList(db:TDatabase; const tablename : string;
  slist : TStrings);
var
  i : integer;
  rtab : TTable;
begin
  slist.BeginUpdate;
  rtab := TTable.Create(nil);
  try
    slist.Clear;
    rtab.DatabaseName := db.DatabaseName;
    rtab.TableName := tablename;
    rtab.FieldDefs.Update;
    WITH rtab.FieldDefs do
    begin
      for i := 0 to Count-1 do
      begin
        slist.Add(Items[i].Name);
      end;
    end;
  finally
    slist.EndUpdate;
    rtab.free;
  end;

end;


procedure CreateFieldsFromDefs(inTable:TDataset);
var
  i : integer;
  def : TFieldDef;
  defs : TFieldDefs;
begin
  defs := inTable.FieldDefs;
  for i := 0 to defs.Count-1 do
  begin
    def := TFieldDef(defs[i]);
    def.CreateField(inTable);
  end;
end;

function FileNameMinusExtension(const inStr : string):string;
var
	i : integer;
begin
	for i := Length(inStr) downto 1 do begin
    	if (inStr[i]='.') then begin
        	Result := Copy(inStr,1,i-1);
            exit;
           	end;
        end;
    Result := inStr;
end;

procedure DBRenameTable(inDB : TDatabase;const oldName, newName : string);
begin
	DbiRenameTable(inDB.Handle, StrToPChar(oldName), nil, StrToPchar(newName));
end;

procedure PackParadoxTable(tabdb : TDatabase; const inTabName : string);
var
	TblDesc: CRTblDesc;
begin
	FillChar(TblDesc, SizeOf(CRTblDesc), 0);
	with TblDesc do begin
		StrPCopy(szTblName, inTabName);
		bPack := True;
	end;
	Check(DbiDoRestructure(tabdb.Handle, 1, @TblDesc, nil, nil, nil, FALSE));
end;

procedure DBPackTable(inDB : TDatabase; inTable : TTable);
begin
 	Check(DBiPackTable(inDB.Handle, inTable.Handle, nil, nil, TRUE));
end;
(*
function DbiPackTable (hDb: hDBIDb; hCursor:

 hDBICur; pszTableName: PChar; pszDriverType: PChar; bRegenId
xs: Bool): DBIResult stdcall;

function DbiPackTable (                 { Pack a table }
      hDb           : hDBIDb;           { Database handle }
      hCursor       : hDBICur;          { Cursor (OR) }
      pszTableName  : PChar;            { Table name }
      pszDriverType : PChar;            { Driver type /NULL }
      bRegenIdxs    : Bool              { Regenerate indexes }
   ): DBIResult;
*)
Procedure UpgradeLevel(inDB : TDatabase; const inName : string);
{table is assumed to be lower level table which will be
up-leveled by a simple copy process}
var
	rtab : TTable;
    ntab : TTable;
    bat : TBatchMove;
    extension : string[31];
  	tempname : string;
    path : string;
    tempNo : LongInt;
    sql : TStrings;

    function GetTempName(inNo : LongInt; const ext : string) : string;
    begin
        Result := Format('T_%6.6d',[inNo]);
        if Length(ext) > 0 then begin
            if ext[1] <> '.' then
            	Result := Result + '.';
            Result := Result + ext;
        	end;
    end;

begin
    path := GetDBDirectory(inDB);
    extension := extractFileExt(inName);
    tempNo := 0;
    tempname := GetTempName(tempNo,extension);


    while FileExists(path+'\'+tempname) do begin
    	 Inc(tempNo);
         if (tempNo > 999999) then
         	 raise EDatabaseError.Create('No temporary name available');
         tempname := GetTempName(tempNo,extension);
         end;

    rtab := TTable.Create(nil);
    ntab := TTable.Create(nil);
    bat := TBatchMove.Create(nil);
    sql := TStringList.Create;
TRY

    AppendCreateTableSQL(inDb, inName,sql);

    DBRenameTable(inDB, inName, tempname);

	ExecuteBDESQL_StrList(inDB.Handle,sql);

    rtab.TableName := tempname;
    rtab.DatabaseName := inDb.DatabaseName;

 	 ntab.TableName := inName;
    ntab.DatabaseName := inDB.DatabaseName;

    bat.Source := rtab;
    bat.Destination := ntab;
    bat.Mode := batAppend;
    bat.Execute;

    DeleteDBTable(inDB, tempname);
FINALLY
	rtab.Free;
    ntab.Free;
    bat.Free;
    sql.Free;
END;

end;

function BDEToDelphiDateTime(inDate:double) : TDateTime;
var
  bdeDate : DbiDate;
  bdeTime : Time;
  iMon, iDay:WORD;
  iYear : SmallInt;
  iHour : WORD;
  iMin : WORD;
  iMilSec : WORD;
  iSec : WORD;
begin
  if inDate = 0 then
  begin {special value}
    Result := 0;
    exit;
  end;
  DbiTimeStampDecode(inDate,bdeDate,bdeTime);
  DbiDateDecode(bdeDate,iMon, iDay, iYear);
  DbiTimeDecode(bdeTime,iHour,iMin,iMilSec);
  iSec := iMilSec div 1000;
  iMilSec := iMilSec - iSec*1000;
  Result := EncodeDate(iYear, iMon, iDay) + EncodeTime(iHour,iMin,iSec,iMilSec);
end;

function DelphiDateTimeToBDE(inDateTime:TDateTime) : double;
var
  bdeDate : DbiDate;
  bdeTime : Time;
  iMon, iDay:WORD;
  iYear : WORD;
  iHour, iMin, iMilSec : WORD;
  iSec : WORD;
begin
  Result := 0;
  DecodeDate(inDateTime,iYear,iMon,iDay);
  DecodeTime(inDateTime,iHour,iMin,iSec,iMilSec);
  iMilsec := iMilsec + iSec * 1000;
  DbiTimeEncode(iHour,iMin, iMilsec,bdeTime);
  DbiDateEncode(iMon,iDay,iYear,bdeDate);
  DbiTimeStampEncode(bdeDate,bdeTime,Result);
end;

Function GetTableLevel(inDB : TDatabase; const inTabName : string) : WORD;
var
	rtab : TTable;
    prop : CURProps;
begin
	rtab := DBOpenTable(inDB, inTabName);
TRY
    DbiGetCursorProps(rtab.Handle,prop);

    Result := prop.iTblLevel;

FINALLY
	rtab.Free;
END;

end;

(*procedure ApplyDatasetToTable(inDS : TRyDataset;
  const inTable : TTable;
  fieldListCopy : TStrings;
  bmode);
var
  vBatch : TBatchMove;
  wasActive : boolean;
begin
  vBatch := nil;
  wasActive := inDS.Active;
  if not wasActive then
    inDS.Active := TRUE;

  if vmode = cvm_empty_append then
    inTable.EmptyTable;
  TRY
    vBatch := TBatchMove.Create(nil);
	   vBatch.Source := inDS;
    vBatch.AbortOnKeyViol := FALSE;
    vBatch.KeyViolTableName := 'KeyViol.DB';
    vBatch.Destination := inTable;
    if (fieldListCopy <> nil) then
      vBatch.Mappings := fieldListCopy;
    vBatch.Mode := batAppend;
    vBatch.Execute;
  FINALLY
    inDS.Active := wasActive;
    vBatch.Free;
  END;
end;*)

function CloneTTable(srcTable : TTable; destDB : TDatabase;
					const destname: string) : TTable;
var
  wasActive : boolean;
begin
  Result := nil;
  wasActive := srcTable.Active;
  if not wasActive then
    srcTable.Active := TRUE;
  TRY
    if DBTableExists(destDB,destname) then
      DeleteDBTable(destDB,destname);
    TRY
      Result := CloneTableStruct(srcTable);
      Result.TableName := destname;
      Result.DatabaseName := destDB.DatabaseName;
      Result.CreateTable;
      ApplyDatasetToTable(srcTable,Result,nil,batAppend,FALSE);
    EXCEPT
      FreeAndNil(TObject(Result));
      raise;
    END;
  FINALLY
    srcTable.Active := wasActive;
  END;
end;

procedure OpenDatabasePath(inDb:TDatabase;const inPath : string);
begin
  with inDb do
  begin
	Connected := FALSE;
    	Params.Clear;
	Params.Add('PATH='+inPath);
	Connected := TRUE;
  end;
end;

function ConstructCreateTableSQL(const tableName: string;
				 const fields: array of TQ_Field;
				 const fieldtypes: array of TQ_Type;
				 numfields,
				 numPrimary: integer): string; forward;

function StrArrayToStr(const inSQL: array of string): string;
var
  i: integer;
begin
  Result := '';
  for i := Low(inSQL) to High(inSQL) do
  begin
    if  i <> Low(inSQL) then Result := Result + ' ';
    Result := Result + inSQL[i];
  end;
end;

function TableHasIndexName(inTable : TTable; const inName:string):Boolean;
var
   i : integer;
   idx : TIndexDef;
begin
	 WITH inTable do begin
		  IndexDefs.Update;
		  for i := 0 to IndexDefs.Count-1 do begin
			  idx := IndexDefs[i];
			  if idx.Name = inName then begin
                 Result := TRUE;
				 exit;
				 end;
			  end;
          end;
	 Result := FALSE;
end;

(*function  OpenDBTable(inDB:TDatabase;const inName:string):TTable;
begin
	Result := TTable.Create(inDB);
    TRY
    Result.TableName := inName;
    Result.DatabaseName := inDb.DatabaseName;
    Result.Open;
    EXCEPT
    on exception do begin
    	Result.Free;
        raise;
        end;
    END;

end; *)


procedure DisposeIndexName(inTable : TTable; const inName : string);
var
   hadIndex : Boolean;
   tempStr : string;
   i : integer;
begin
	 hadIndex := TableHasIndexName(inTable,inName);
	 tempStr := inTable.TableName;

	 i := Pos('.',tempStr);
	 if ( i > 0) then
		tempStr := copy(tempStr,1,i-1);
	 if hadIndex then
		ExecuteDatabaseSQL(inTable.Database,'Drop index '+tempStr+'.'+inName);
end;

procedure ExecuteCreateTableSQL(inDB: TDatabase; const intablename : string;
                                                 const InArray: array of string);
var
   slist : TStringList;
   scount : integer;
   j,k : integer;
   db_h : HDBIDB;

begin
   j := Low(InArray);
   k := High(InArray);

   scount := k - j + 1;

   if scount <= 0 then
     raise EDBToolsException.Create('Empty string array');


   db_h := inDB.Handle;
  if (db_h = nil) then
     raise EDBToolsException.Create('Nil database handle');

   slist := TStringList.Create;
TRY
   slist.Add(Format( InArray[j], [intablename]) );
   Inc(j);
   while( j <= k ) do begin
      slist.Add( InArray[j]) ;
      Inc(j);
      end;
   ExecuteBDESQL_StrList(db_h, slist);
FINALLY
   slist.Free;
END;
end;

function DBOpenQuery(inDB : TDatabase; const sql : string) : TQuery;
begin
  Result := TQuery.Create(nil);
  try
    Result.DatabaseName := inDB.DatabaseName;
    Result.SQL.Add(sql);
    Result.Open;
  EXCEPT
    Result.Free;
    raise;
  END;
end;


function  DBOpenTable(inDB:TDatabase; const tableName: string) : TTable;
begin

	Result := TTable.Create(nil);
	TRY
	Result.DatabaseName := inDB.DatabaseName;
	Result.TableName := tableName;
	Result.Open;
	EXCEPT
	On Exception do begin
		Result.Free;
		raise;
		end;
	END;

end;

procedure DeleteDBTable(inDB: TDatabase; const tableName: string);
begin
  DbiDeleteTable(inDB.Handle, StrToPChar(tableName), nil);
end;

function DBTableExists(inDB: TDatabase; const tableName: string): boolean;
var
  List: TStringList;
  name, noext : string;
  i : integer;
begin
  List := nil;
  try
    ExtractFileNameExt(tableName,name,noext);
    List := TStringList.Create;
    Session.GetTableNames(inDB.DatabaseName, '*.*', FALSE, FALSE, List);
    Result := FALSE;
    for i := 0 to List.Count-1 do
    begin
      if (CompareText(name,List[i])=0) then
      begin
        Result := TRUE;
        break;
      end;
    end;
  finally
    List.Free;
  end;
end;

procedure SetListStrArray(slist: TStrings; const inArray : array of string);
var
  i: integer;
begin
  slist.Clear;
  for i := Low(inArray) to High(inArray) do slist.Add(inArray[i]);
end;


procedure ExecuteQuerySQL(inQuery: TQuery; const InArray : array of string);
begin
  with inQuery do
  begin
    Active := FALSE;
    SetListStrArray(SQL, InArray);
    Active := TRUE;
  end;
end;

function ApplyDatasetToTable(
  inQuery: TRyDataset; const inTable: TTable;
  fieldListCopy: TStrings; bmode: TBatchMode; emptytable:boolean) : integer;
{If fieldListCopy is nil, all the fields in the query are copied to
the destination table}
var
  vBatch: TBatchMove;
  slist: TStrings;
  i: integer;
  s : string;
begin
  vBatch := nil;
  if (inQuery.Active = FALSE) then
  begin
    s := Format('Source Query for %s not active',[inTable.TableName]);
    raise EDatabaseError.Create(s);
  end;
  if emptytable then
    inTable.EmptyTable;
  if fieldListCopy = nil then
  begin
    slist := TStringList.Create;
    for i := 0 to inQuery.FieldCount-1 do
    begin
      s := inQuery.Fields[i].FieldName;
      s := Format('%s=%s',[s,s]);
      slist.Add(s);
    end;
  end
  else
    slist := fieldListCopy;

  try
    vBatch := TBatchMove.Create(nil);
    vBatch.Source := inQuery;
    vBatch.AbortOnKeyViol := FALSE;
    vBatch.KeyViolTableName := 'KeyViol.DB';
    vBatch.ProblemTableName := 'Problem.DB';
    vBatch.Destination := inTable;
    vBatch.Mappings := slist;
    vBatch.Mode := batAppend;
    vBatch.Execute;
    Result := vBatch.MovedCount;
    WITH vBatch do
    begin
      if (ProblemCount>0) or (KeyViolCount>0) then
      begin
        s := Format('Batch to %s failed. %d problem records %d key violations',
          [inTable.TableName,ProblemCount,KeyViolCount]);
        raise EDatabaseError.Create(s);
      end;
    end;
  finally
    if fieldListCopy = nil then
      slist.Free;
    vBatch.Free;
  end;
end;

function GetSQLFieldDef(fieldDef : TFielddef) : string;

function sizeStr : string;
begin
  REsult := IntToStr(fieldDef.Size);
end;

begin
  case fieldDef.DataType of
    ftString:     Result := 'VARCHAR('+sizeStr+')';
    ftSmallInt:	  Result := 'SMALLINT';
    ftInteger:    Result := 'INTEGER';
    ftFloat:   	  Result := 'NUMERIC';
    ftDate :      Result := 'DATE';
    ftTime :      Result := 'TIME';
    ftDateTime:   Result := 'TIMESTAMP';
    ftBoolean:    Result := 'BOOLEAN';
  end; { end case }
  Result := fieldDef.Name + ' '+Result;
end;

procedure FillStringsFromQuery(
  inDB : TDatabase;
  const sql : string;
  slist : TStrings);
var
  qry : TQuery;
begin
  qry := TQuery.Create(nil);
  slist.BeginUpdate;
  try
    slist.Clear;
    qry.DatabaseName := inDB.DatabaseName;
    qry.SQL.Add(sql);
    qry.Open;

    while not qry.eof do
    begin
      slist.Add(qry.Fields[0].AsString);
      qry.next;
    end;
  finally
    slist.EndUpdate;
    qry.free;
  end;
end;

Function InstantiateQueryDB(inQuery : TQuery;
  destDB : TDatabase;
  const inTableName : string;
  keyCount : integer) : TTable;
var
  i : integer;
  tstr : string;
begin
  inQuery.Active := TRUE;
  Result := TTable.Create(nil);
  TRY
    Result.TableName := inTableName;
    if DBTableExists(destDB,inTableName) then
      DeleteDBTable(inQuery.Database,inTableName);
      
    Result.DatabaseName := destDB.DatabaseName;
    Result.FieldDefs.Assign(inQuery.FieldDefs);
    tstr := '';
    if keycount > 0 then
    begin
      for i := 0 to keyCount-1 do
      begin
        if i > 0 then
        tstr := tstr + ';';
        tstr := tstr + UPPERCASE(Result.FieldDefs[i].Name);
      end;
      Result.IndexDefs.Add('',tstr,[ixPrimary]);
    end;
    Result.CreateTable;
    ApplyDatasetToTable(inQuery,Result,nil,batAppend,FALSE);
  EXCEPT
    FreeAndNil(TObject(Result));
    raise;
  END;
end;

Function InstantiateQuery(inQuery : TQuery;
							const inTableName : string;
							keyCount : integer) : TTable;
var
  i : integer;
  tstr : string;

begin
  inQuery.Active := TRUE;
  Result := TTable.Create(nil);

  TRY
    Result.TableName := inTableName;
    Result.DatabaseName := inQuery.DatabaseName;
    if DBTableExists(inQuery.Database,inTableName) then
    DeleteDBTable(inQuery.Database,inTableName);
    Result.FieldDefs.Assign(inQuery.FieldDefs);
    tstr := '';
    if keycount > 0 then
    begin
      for i := 0 to keyCount-1 do
      begin
        if i > 0 then
        tstr := tstr + ';';
        tstr := tstr + UPPERCASE(Result.FieldDefs[i].Name);
      end;
      Result.IndexDefs.Add('',tstr,[ixPrimary]);
    end;
    Result.CreateTable;
    ApplyDatasetToTable(inQuery,Result,nil,batAppend,FALSE);

  EXCEPT
    FreeAndNil(TObject(Result));
    raise;
  END;
end;

function QueryAsTable(
  inQuery: TQuery; const inTableName: string;
  fieldListCopy: TStrings; emptytable : boolean): TTable;
begin
  if inQuery.Active = FALSE then
    raise EDatabaseError.Create('Query '+inQuery.Name+' not active');
  Result := TTable.Create(Application);
  try
    Result.DatabaseName := inQuery.DatabaseName;
    Result.TableName := inTableName;
    ApplyDatasetToTable(inQuery, Result, fieldListCopy, batAppend, emptytable);
  except
    FreeAndNil(TObject(Result));
    raise;
  end;
end;


procedure DbiError(ErrorCode: DBIResult);
begin
  raise EDBEngineError.Create(ErrorCode);
end;

procedure Check(Status: DBIResult);
begin
  if Status <> 0 then DbiError(Status);
end;

function ConstructCreateTableSQL(const tableName: string;
  				 const fields: array of TQ_Field;
				 const fieldtypes: array of TQ_Type;
				 numfields,
				 numPrimary: integer): string;
var
  i: integer;
  tempStr: String[255];
begin
  Result := 'Create Table "'+tableName + '" (';
  for i := 0 to numfields-1 do
  begin
    if (i > 0) then
      tempStr := ','
	else
	  tempStr := '';
	tempStr := tempStr +  fields[i]+' '+fieldtypes[i];
    Result := Result + tempStr;
  end;

  if (numPrimary > 0) then
  begin
    tempStr := '';
    for i := 0 to numPrimary-1 do
    begin
      if i > 0 then tempStr := tempStr + ',';
      tempStr := tempStr +  fields[i];
    end;
    Result := Result + ', Primary Key(' + tempStr + ')';
  end;

  Result := Result + ')';
end;

procedure ExecuteBDESQL_StrList(hDB: HDBIDB; slist : TStrings);
var
  hStmt: HDBIStmt;
  hCur : HDBICur;
  pc : PChar;
begin
  hStmt := nil;
  hCur := nil;
  pc := StrListToCStr(slist,' ');
  TRY
  try

{$IFDEF VER80}
	Check(DbiQPrepare(hDB, qrylangSQL, pc, hStmt));

{$ELSE}
	Check(DbiQAlloc(hDB, qrylangSQL, hStmt));
    Check(DbiQPrepare(hStmt,pc));

{$ENDIF}
    Check(DbiQExec(hStmt, @hCur));
    EXCEPT
{$IFDEF DEBUG_LOG}
	if gLogOpen then
		Writeln(gLog, pc);
    CloseDebugLog;
{$ENDIF}
    END;

  finally
	StrDispose(pc);
    if hStmt <> nil then Check(DbiQFree(hStmt));
    if hCur <> nil then Check(DbiCloseCursor(hCur));
  end;
end;

procedure ExecuteBDESQL(hDB: HDBIDB; const sql :  string);
var
  hStmt: HDBIStmt;
  hCur : HDBICur;
  pc : PChar;
begin
  hStmt := nil;
  hCur := nil;
  pc := StrToPChar(sql);
  try

{$IFDEF VER80}
    Check(DbiQPrepare(hDB, qrylangSQL, pc, hStmt));

{$ELSE}
    Check(DbiQAlloc(hDB, qrylangSQL, hStmt));
    Check(DbiQPrepare(hStmt,pc));

{$ENDIF}
    Check(DbiQExec(hStmt, @hCur));

  finally
    if hStmt <> nil then Check(DbiQFree(hStmt));
    if hCur <> nil then Check(DbiCloseCursor(hCur));
  end;
end;

procedure ExecuteAliasSQL(const aliasDB : string; const sql :  string);
var
  hDB : HDBIDB;
begin
  try
    hDB := nil;
	Check(DbiOpenDatabase(StrToPChar(aliasDB),nil,dbiREADWRITE, dbiOPENSHARED,nil,0,nil,nil,hDB));
	ExecuteBDESQL(hDB, sql);
  finally
    if hDB <> nil then DbiCloseDatabase(hDB);
  end;
end;

function GetDBDriverName(db:TDatabase):string;
{$IFDEF VER80}
var  plist : TStrings;
{$ENDIF}
begin
{$IFDEF VER80}
  plist := TStringList.create;
  TRY
    Session.GetAliasParams(db.DatabaseName, plist);
    Result := plist.Values['TYPE'];
  FINALLY
    plist.free;
  END;
{$ELSE}
  Result := db.Session.GetAliasDriverName(db.DatabaseName);
{$ENDIF}
end;

procedure DeleteTableSQL(const alias, tablename : string);
var
  delstr : string;
begin
  delstr := 'DROP TABLE '''+tablename+'''';
  ExecuteAliasSQL(alias,delstr);
end;

procedure ExecuteDatabaseSQL(inDB: TDatabase; const sql :  string);
var
   db_h : HDBIDB;
begin
  db_h := inDB.Handle;
  if (db_h = nil) then
     raise EDBToolsException.Create('Nil database handle');
  ExecuteBDESQL(db_h,sql);
end;

procedure CreateTableSQL(const alias, tableName: string;
			 const fields: array of TQ_Field;
			 const fieldtypes: array of TQ_Type;
			 numfields, numPrimary: integer
			 );
var
  sql: string;
begin
  sql := ConstructCreateTableSQL(tableName, fields, fieldtypes, numfields, numPrimary);
  ExecuteAliasSQL(alias,sql);
end;

function  EncloseInQuotes(const inStr : string) : string;
var
  i,rq,k : integer;
const
  kSQUOTE = '''';
begin
  {find out the size of result, duplicate internal quotes}
  rq := 0;
  for i := 1 to Length(inStr) do
    if inStr[i] = kSQUOTE then
      Inc(rq);
  rq := rq + Length(inStr)+2;
  SetLength(Result,rq);
  Result[1] := kSQUOTE;  {start quote}
  Result[rq] := kSQUOTE;  {end quote}
  k := 1;
  for i := 1 to Length(inStr) do
  begin
    if inStr[i] <> kSQUOTE then
    begin
      Inc(k);
      Result[k] := inStr[i];
    end else
    begin
      Inc(k);
      Result[k] := kSQUOTE;
      Inc(k);
      Result[k] := kSQUOTE;
    end;
  end;
end;

function QuoteCommaList(slist: TStrings): string;
var
  i: integer;
begin
  Result := '';
  for i := 0 to slist.Count-1 do
  begin
    if i > 0 then
      Result := Result + ',';
    Result := Result + EncloseInQuotes(slist[i]);
  end;
end;

end.
