{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnUtIB.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnUtIB;

interface
Uses Classes, IBDatabase, IBSQL, SysUtils, IBCustomDataset, IBQuery;

type
  ERnnUtIB = class(Exception);


procedure ConnectIBDatabase(db : TIBDatabase;
   const path, user, pw, role : string);

function GetRole(db : TIBDatabase) : string;
function GetUser(db : TIBDatabase) : string;

function CreateIBSQL(
  trans : TIBTransaction;
  db : TIBDatabase;
  const sql : string;
  const args : Variant) : TIBSQL;

function CreateIBDataset(
  trans : TIBTransaction;
  db : TIBDatabase;
  const sql : string;
  const args : Variant) : TIBDataset;

procedure FillStringsFromIBDataset(
  inDB : TIBDatabase;
  const sql : string;
  slist : TStrings);

procedure SetAllDatabaseTrans(dm : TComponent;
  db : TIBDatabase; trans : TIBTransaction);

  
implementation
const
  kusername = 'user_name';
  ksqlrole = 'sql_role';
  kpassword = 'password';

function GetRole(db : TIBDatabase) : string;
begin
  Result := db.params.Values[ksqlrole];
end;

function GetUser(db : TIBDatabase) : string; 
begin
  Result := db.params.Values[kusername];
end;

procedure SetAllDatabaseTrans(dm : TComponent;
  db : TIBDatabase; trans : TIBTransaction);
var
  i : integer;
  comp : TComponent;
begin
  with dm do
  for i := 0 to ComponentCount-1 do
  begin
    comp := Components[i];
    if comp is TIBSQL then
    with comp as TIBSQL do
    begin
      Database := db;
      Transaction := trans;
    end else
    if comp is TIBCustomDataset then
    with comp as TIBCustomDataset do
    begin
      Database := db;
      Transaction := trans;
    end else
    if comp is TIBDataset then
    with comp as TIBDataset do
    begin
      Database := db;
      Transaction := trans;
    end
    else if comp is TIBQuery then
    with comp as TIBQuery do
    begin
      Database := db;
      Transaction := trans;
    end;
  end;
end;

procedure ConnectIBDatabase(db : TIBDatabase;
   const path, user, pw, role : string);
var
  ix : integer;
begin
  if db.Connected then
  begin
    raise ERnnUtIB.Create('Database Already connected');
    exit;
  end;
  db.DatabaseName := path;
  if (length(path) < 0) or (length(user) < 0) then
    db.LoginPrompt := TRUE
  else
    db.LoginPrompt := FALSE;
  with db.Params do
  begin
    Values[kusername] := user;
    Values[kpassword] := pw;
    if (Length(role) > 0) then
    begin
       Values[ksqlrole] := pw;
    end else
    begin
      ix := IndexOf(ksqlrole);
      if ix > 0 then
        Delete(ix);
    end;
  end;
  db.Connected := TRUE;
end;



function CreateIBSQL(
  trans : TIBTransaction;
  db : TIBDatabase;
  const sql : string;
  const args : Variant) : TIBSQL;
var
  i : integer;
  below, uphigh : integer;
begin
  Result := TIBSQL.Create(nil);
  try
    Result.Transaction := trans;
    Result.Database := db;
    Result.SQL.Text := sql;
    Result.Prepare;
    if VarIsArray(args) then
    begin
      below := VarArrayLowBound(args,1);
      uphigh := VarArrayHighBound(args,1);
      for i := below to uphigh do
        Result.Params[i-below].AsVariant := args[i];
    end else
    if not VarIsEmpty(args) then
      Result.Params[0].AsVariant := args;
    Result.ExecQuery;
  except
    Result.Free;
  end;
end;

function CreateIBDataset(
  trans : TIBTransaction;
  db : TIBDatabase;
  const sql : string;
  const args : Variant) : TIBDataset;
{create dataset with at select sql, given the parameters}
var
  i : integer;
  below, uphigh : integer;
begin
  Result := TIBDataset.Create(nil);
  try
    Result.Transaction := trans;
    Result.Database := db;
    Result.SelectSQL.Text := sql;
    Result.Prepare;
    if VarIsArray(args) then
    begin
      below := VarArrayLowBound(args,1);
      uphigh := VarArrayHighBound(args,1);
      for i := below to uphigh do
        Result.Params[i-below].AsVariant := args[i];
    end else
    if not VarIsEmpty(args) then
      Result.Params[0].AsVariant := args;
  except
    Result.Free;
  end;
end;

procedure FillStringsFromIBDataset(
  inDB : TIBDatabase;
  const sql : string;
  slist : TStrings);
var
  qry : TIBDataset;
begin
  qry := TIBDataset.Create(nil);
  slist.BeginUpdate;
  try
    slist.Clear;
    qry.Database := inDB;
    qry.SelectSQL.Text := sql;
    qry.Open;

    while not qry.eof do
    begin
      slist.Add(TRIM(qry.Fields[0].AsString));
      qry.next;
    end;
  finally
    slist.EndUpdate;
    qry.free;
  end;
end;

end.
